from __future__ import annotations

import re
import sys
from asyncio import run
from collections.abc import Callable
from dataclasses import dataclass
from functools import partial
from getpass import getuser
from itertools import repeat
from pathlib import Path
from socket import gethostname
from traceback import TracebackException
from typing import TYPE_CHECKING, override

from utilities.datetime import get_datetime, get_now, serialize_compact
from utilities.errors import repr_error
from utilities.iterables import OneEmptyError, one
from utilities.pathlib import get_path
from utilities.reprlib import (
    RICH_EXPAND_ALL,
    RICH_INDENT_SIZE,
    RICH_MAX_DEPTH,
    RICH_MAX_LENGTH,
    RICH_MAX_STRING,
    RICH_MAX_WIDTH,
    yield_mapping_repr,
)
from utilities.version import get_version
from utilities.whenever import serialize_duration

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator, Sequence
    from traceback import FrameSummary
    from types import TracebackType

    from utilities.types import MaybeCallableDateTime, MaybeCallablePathLike, PathLike
    from utilities.version import MaybeCallableVersionLike


##


_START = get_now()


def format_exception_stack(
    error: BaseException,
    /,
    *,
    header: bool = False,
    start: MaybeCallableDateTime | None = _START,
    version: MaybeCallableVersionLike | None = None,
    capture_locals: bool = False,
    max_width: int = RICH_MAX_WIDTH,
    indent_size: int = RICH_INDENT_SIZE,
    max_length: int | None = RICH_MAX_LENGTH,
    max_string: int | None = RICH_MAX_STRING,
    max_depth: int | None = RICH_MAX_DEPTH,
    expand_all: bool = RICH_EXPAND_ALL,
) -> str:
    """Format an exception stack."""
    lines: Sequence[str] = []
    if header:
        lines.extend(_yield_header_lines(start=start, version=version))
    lines.extend(
        _yield_formatted_frame_summary(
            error,
            capture_locals=capture_locals,
            max_width=max_width,
            indent_size=indent_size,
            max_length=max_length,
            max_string=max_string,
            max_depth=max_depth,
            expand_all=expand_all,
        )
    )
    return "\n".join(lines)


def _yield_header_lines(
    *,
    start: MaybeCallableDateTime | None = _START,
    version: MaybeCallableVersionLike | None = None,
) -> Iterator[str]:
    """Yield the header lines."""
    from utilities.tzlocal import get_local_time_zone, get_now_local
    from utilities.whenever import serialize_zoned_datetime

    now = get_now_local()
    start_use = get_datetime(datetime=start)
    start_use = (
        None if start_use is None else start_use.astimezone(get_local_time_zone())
    )
    yield f"Date/time | {serialize_zoned_datetime(now)}"
    start_str = "" if start_use is None else serialize_zoned_datetime(start_use)
    yield f"Started   | {start_str}"
    duration = None if start_use is None else (now - start_use)
    duration_str = "" if duration is None else serialize_duration(duration)
    yield f"Duration  | {duration_str}"
    yield f"User      | {getuser()}"
    yield f"Host      | {gethostname()}"
    version_use = "" if version is None else get_version(version=version)
    yield f"Version   | {version_use}"
    yield ""


def _yield_formatted_frame_summary(
    error: BaseException,
    /,
    *,
    capture_locals: bool = False,
    max_width: int = RICH_MAX_WIDTH,
    indent_size: int = RICH_INDENT_SIZE,
    max_length: int | None = RICH_MAX_LENGTH,
    max_string: int | None = RICH_MAX_STRING,
    max_depth: int | None = RICH_MAX_DEPTH,
    expand_all: bool = RICH_EXPAND_ALL,
) -> Iterator[str]:
    """Yield the formatted frame summary lines."""
    stack = TracebackException.from_exception(
        error, capture_locals=capture_locals
    ).stack
    n = len(stack)
    for i, frame in enumerate(stack, start=1):
        num = f"{i}/{n}"
        first, *rest = _yield_frame_summary_lines(
            frame,
            max_width=max_width,
            indent_size=indent_size,
            max_length=max_length,
            max_string=max_string,
            max_depth=max_depth,
            expand_all=expand_all,
        )
        yield f"{num} | {first}"
        blank = "".join(repeat(" ", len(num)))
        for rest_i in rest:
            yield f"{blank} | {rest_i}"
    yield repr_error(error)


def _yield_frame_summary_lines(
    frame: FrameSummary,
    /,
    *,
    max_width: int = RICH_MAX_WIDTH,
    indent_size: int = RICH_INDENT_SIZE,
    max_length: int | None = RICH_MAX_LENGTH,
    max_string: int | None = RICH_MAX_STRING,
    max_depth: int | None = RICH_MAX_DEPTH,
    expand_all: bool = RICH_EXPAND_ALL,
) -> Iterator[str]:
    module = _path_to_dots(frame.filename)
    yield f"{module}:{frame.lineno} | {frame.name} | {frame.line}"
    if frame.locals is not None:
        yield from yield_mapping_repr(
            frame.locals,
            _max_width=max_width,
            _indent_size=indent_size,
            _max_length=max_length,
            _max_string=max_string,
            _max_depth=max_depth,
            _expand_all=expand_all,
        )


def _path_to_dots(path: PathLike, /) -> str:
    new_path: Path | None = None
    for pattern in [
        "site-packages",
        ".venv",  # after site-packages
        "src",
        r"python\d+\.\d+",
    ]:
        if (new_path := _trim_path(path, pattern)) is not None:
            break
    path_use = Path(path) if new_path is None else new_path
    return ".".join(path_use.with_suffix("").parts)


def _trim_path(path: PathLike, pattern: str, /) -> Path | None:
    parts = Path(path).parts
    compiled = re.compile(f"^{pattern}$")
    try:
        i = one(i for i, p in enumerate(parts) if compiled.search(p))
    except OneEmptyError:
        return None
    return Path(*parts[i + 1 :])


##


def make_except_hook(
    *,
    start: MaybeCallableDateTime | None = _START,
    version: MaybeCallableVersionLike | None = None,
    path: MaybeCallablePathLike | None = None,
    max_width: int = RICH_MAX_WIDTH,
    indent_size: int = RICH_INDENT_SIZE,
    max_length: int | None = RICH_MAX_LENGTH,
    max_string: int | None = RICH_MAX_STRING,
    max_depth: int | None = RICH_MAX_DEPTH,
    expand_all: bool = RICH_EXPAND_ALL,
    slack_url: str | None = None,
) -> Callable[
    [type[BaseException] | None, BaseException | None, TracebackType | None], None
]:
    """Exception hook to log the traceback."""
    return partial(
        _make_except_hook_inner,
        start=start,
        version=version,
        path=path,
        max_width=max_width,
        indent_size=indent_size,
        max_length=max_length,
        max_string=max_string,
        max_depth=max_depth,
        expand_all=expand_all,
        slack_url=slack_url,
    )


def _make_except_hook_inner(
    exc_type: type[BaseException] | None,
    exc_val: BaseException | None,
    traceback: TracebackType | None,
    /,
    *,
    start: MaybeCallableDateTime | None = _START,
    version: MaybeCallableVersionLike | None = None,
    path: MaybeCallablePathLike | None = None,
    max_width: int = RICH_MAX_WIDTH,
    indent_size: int = RICH_INDENT_SIZE,
    max_length: int | None = RICH_MAX_LENGTH,
    max_string: int | None = RICH_MAX_STRING,
    max_depth: int | None = RICH_MAX_DEPTH,
    expand_all: bool = RICH_EXPAND_ALL,
    slack_url: str | None = None,
) -> None:
    """Exception hook to log the traceback."""
    _ = (exc_type, traceback)
    if exc_val is None:
        raise MakeExceptHookError
    slim = format_exception_stack(exc_val, header=True, start=start, version=version)
    _ = sys.stderr.write(f"{slim}\n")  # don't 'from sys import stderr'
    if path is not None:
        from utilities.atomicwrites import writer
        from utilities.tzlocal import get_now_local

        path = (
            get_path(path=path)
            .joinpath(serialize_compact(get_now_local()))
            .with_suffix(".txt")
        )
        full = format_exception_stack(
            exc_val,
            header=True,
            start=start,
            version=version,
            capture_locals=True,
            max_width=max_width,
            indent_size=indent_size,
            max_length=max_length,
            max_string=max_string,
            max_depth=max_depth,
            expand_all=expand_all,
        )
        with writer(path, overwrite=True) as temp:
            _ = temp.write_text(full)
    if slack_url is not None:  # pragma: no cover
        from utilities.slack_sdk import send_to_slack

        send = f"```{slim}```"
        run(send_to_slack(slack_url, send))


@dataclass(kw_only=True, slots=True)
class MakeExceptHookError(Exception):
    @override
    def __str__(self) -> str:
        return "No exception to log"


__all__ = ["format_exception_stack", "make_except_hook"]



from __future__ import annotations

from typing import TYPE_CHECKING

from hypothesis import given, settings
from pytest import raises

from utilities.git import GetRepoRootError, get_repo_root
from utilities.hypothesis import git_repos

if TYPE_CHECKING:
    from pathlib import Path


class TestGetRepoRoot:
    @given(repo=git_repos())
    @settings(max_examples=1)
    def test_main(self, *, repo: Path) -> None:
        root = get_repo_root(path=repo)
        expected = repo.resolve()
        assert root == expected

    def test_error(self, *, tmp_path: Path) -> None:
        with raises(
            GetRepoRootError, match="Path is not part of a `git` repository: .*"
        ):
            _ = get_repo_root(path=tmp_path)



from __future__ import annotations

import datetime as dt
import re
from contextlib import suppress
from dataclasses import dataclass
from functools import cache
from logging import LogRecord
from typing import TYPE_CHECKING, Any, override

from whenever import (
    Date,
    DateTimeDelta,
    PlainDateTime,
    Time,
    TimeZoneNotFoundError,
    ZonedDateTime,
)

from utilities.datetime import (
    _MICROSECONDS_PER_DAY,
    _MICROSECONDS_PER_SECOND,
    ZERO_TIME,
    check_date_not_datetime,
    datetime_duration_to_microseconds,
    parse_two_digit_year,
)
from utilities.math import ParseNumberError, parse_number
from utilities.re import (
    ExtractGroupError,
    ExtractGroupsError,
    extract_group,
    extract_groups,
)
from utilities.zoneinfo import UTC, ensure_time_zone, get_time_zone_name

if TYPE_CHECKING:
    from zoneinfo import ZoneInfo

    from utilities.types import (
        DateLike,
        DateTimeLike,
        Duration,
        DurationLike,
        TimeDeltaLike,
        TimeLike,
    )


MAX_SERIALIZABLE_TIMEDELTA = dt.timedelta(days=3652060, microseconds=-1)
MIN_SERIALIZABLE_TIMEDELTA = -MAX_SERIALIZABLE_TIMEDELTA


##


def check_valid_zoned_datetime(datetime: dt.datetime, /) -> None:
    """Check if a zoned datetime is valid."""
    time_zone = ensure_time_zone(datetime)  # skipif-ci-and-windows
    datetime2 = datetime.replace(tzinfo=time_zone)  # skipif-ci-and-windows
    try:  # skipif-ci-and-windows
        result = (
            ZonedDateTime.from_py_datetime(datetime2)
            .to_tz(get_time_zone_name(UTC))
            .to_tz(get_time_zone_name(time_zone))
            .py_datetime()
        )
    except TimeZoneNotFoundError:  # pragma: no cover
        raise _CheckValidZonedDateTimeInvalidTimeZoneError(datetime=datetime) from None
    if result != datetime2:  # skipif-ci-and-windows
        raise _CheckValidZonedDateTimeUnequalError(datetime=datetime, result=result)


@dataclass(kw_only=True, slots=True)
class CheckValidZonedDateTimeError(Exception):
    datetime: dt.datetime


@dataclass(kw_only=True, slots=True)
class _CheckValidZonedDateTimeInvalidTimeZoneError(CheckValidZonedDateTimeError):
    @override
    def __str__(self) -> str:
        return f"Invalid timezone; got {self.datetime.tzinfo}"  # pragma: no cover


@dataclass(kw_only=True, slots=True)
class _CheckValidZonedDateTimeUnequalError(CheckValidZonedDateTimeError):
    result: dt.datetime

    @override
    def __str__(self) -> str:
        return f"Zoned datetime must be valid; got {self.datetime} != {self.result}"  # skipif-ci-and-windows


##


def ensure_date(date: DateLike, /) -> dt.date:
    """Ensure the object is a date."""
    if isinstance(date, dt.date):
        check_date_not_datetime(date)
        return date
    try:
        return parse_date(date)
    except ParseDateError as error:
        raise EnsureDateError(date=error.date) from None


@dataclass(kw_only=True, slots=True)
class EnsureDateError(Exception):
    date: str

    @override
    def __str__(self) -> str:
        return f"Unable to ensure date; got {self.date!r}"


##


def ensure_datetime(datetime: DateTimeLike, /) -> dt.datetime:
    """Ensure the object is a datetime."""
    if isinstance(datetime, dt.datetime):
        return datetime  # skipif-ci-and-windows
    try:
        return parse_datetime(datetime)
    except ParseDateTimeError as error:
        raise EnsureDateTimeError(datetime=error.datetime) from None


@dataclass(kw_only=True, slots=True)
class EnsureDateTimeError(Exception):
    datetime: str

    @override
    def __str__(self) -> str:
        return f"Unable to ensure datetime; got {self.datetime!r}"


##


def ensure_duration(duration: DurationLike, /) -> Duration:
    """Ensure the object is a Duration."""
    if isinstance(duration, int | float | dt.timedelta):
        return duration
    try:
        return parse_duration(duration)
    except ParseDurationError as error:
        raise EnsureDurationError(duration=error.duration) from None


@dataclass(kw_only=True, slots=True)
class EnsureDurationError(Exception):
    duration: str

    @override
    def __str__(self) -> str:
        return f"Unable to ensure duration; got {self.duration!r}"


##


def ensure_plain_datetime(datetime: DateTimeLike, /) -> dt.datetime:
    """Ensure the object is a plain datetime."""
    if isinstance(datetime, dt.datetime):
        return datetime
    try:
        return parse_plain_datetime(datetime)
    except ParsePlainDateTimeError as error:
        raise EnsurePlainDateTimeError(datetime=error.datetime) from None


@dataclass(kw_only=True, slots=True)
class EnsurePlainDateTimeError(Exception):
    datetime: str

    @override
    def __str__(self) -> str:
        return f"Unable to ensure plain datetime; got {self.datetime!r}"


##


def ensure_time(time: TimeLike, /) -> dt.time:
    """Ensure the object is a time."""
    if isinstance(time, dt.time):
        return time
    try:
        return parse_time(time)
    except ParseTimeError as error:
        raise EnsureTimeError(time=error.time) from None


@dataclass(kw_only=True, slots=True)
class EnsureTimeError(Exception):
    time: str

    @override
    def __str__(self) -> str:
        return f"Unable to ensure time; got {self.time!r}"


##


def ensure_timedelta(timedelta: TimeDeltaLike, /) -> dt.timedelta:
    """Ensure the object is a timedelta."""
    if isinstance(timedelta, dt.timedelta):
        return timedelta
    try:
        return parse_timedelta(timedelta)
    except _ParseTimedeltaParseError as error:
        raise _EnsureTimedeltaParseError(timedelta=error.timedelta) from None
    except _ParseTimedeltaNanosecondError as error:
        raise _EnsureTimedeltaNanosecondError(
            timedelta=error.timedelta, nanoseconds=error.nanoseconds
        ) from None


@dataclass(kw_only=True, slots=True)
class EnsureTimedeltaError(Exception):
    timedelta: str


@dataclass(kw_only=True, slots=True)
class _EnsureTimedeltaParseError(EnsureTimedeltaError):
    @override
    def __str__(self) -> str:
        return f"Unable to ensure timedelta; got {self.timedelta!r}"


@dataclass(kw_only=True, slots=True)
class _EnsureTimedeltaNanosecondError(EnsureTimedeltaError):
    nanoseconds: int

    @override
    def __str__(self) -> str:
        return f"Unable to ensure timedelta; got {self.nanoseconds} nanoseconds"


##


def ensure_zoned_datetime(datetime: DateTimeLike, /) -> dt.datetime:
    """Ensure the object is a zoned datetime."""
    if isinstance(datetime, dt.datetime):
        return datetime
    try:
        return parse_zoned_datetime(datetime)
    except ParseZonedDateTimeError as error:
        raise EnsureZonedDateTimeError(datetime=error.datetime) from None


@dataclass(kw_only=True, slots=True)
class EnsureZonedDateTimeError(Exception):
    datetime: str

    @override
    def __str__(self) -> str:
        return f"Unable to ensure zoned datetime; got {self.datetime!r}"


##


_PARSE_DATE_YYMMDD_REGEX = re.compile(r"^(\d{2})(\d{2})(\d{2})$")


def parse_date(date: str, /) -> dt.date:
    """Parse a string into a date."""
    try:
        w_date = Date.parse_common_iso(date)
    except ValueError:
        try:
            ((year2, month, day),) = _PARSE_DATE_YYMMDD_REGEX.findall(date)
        except ValueError:
            raise ParseDateError(date=date) from None
        year = parse_two_digit_year(year2)
        return dt.date(year=int(year), month=int(month), day=int(day))
    return w_date.py_date()


@dataclass(kw_only=True, slots=True)
class ParseDateError(Exception):
    date: str

    @override
    def __str__(self) -> str:
        return f"Unable to parse date; got {self.date!r}"


##


def parse_datetime(datetime: str, /) -> dt.datetime:
    """Parse a string into a datetime."""
    with suppress(ParsePlainDateTimeError):
        return parse_plain_datetime(datetime)
    with suppress(ParseZonedDateTimeError):
        return parse_zoned_datetime(datetime)
    raise ParseDateTimeError(datetime=datetime) from None


@dataclass(kw_only=True, slots=True)
class ParseDateTimeError(Exception):
    datetime: str

    @override
    def __str__(self) -> str:
        return f"Unable to parse datetime; got {self.datetime!r}"


##


def parse_duration(duration: str, /) -> Duration:
    """Parse a string into a Duration."""
    with suppress(ParseNumberError):
        return parse_number(duration)
    try:
        return parse_timedelta(duration)
    except ParseTimedeltaError:
        raise ParseDurationError(duration=duration) from None


@dataclass(kw_only=True, slots=True)
class ParseDurationError(Exception):
    duration: str

    @override
    def __str__(self) -> str:
        return f"Unable to parse duration; got {self.duration!r}"


##


def parse_plain_datetime(datetime: str, /) -> dt.datetime:
    """Parse a string into a plain datetime."""
    try:
        ldt = PlainDateTime.parse_common_iso(datetime)
    except ValueError:
        raise ParsePlainDateTimeError(datetime=datetime) from None
    return ldt.py_datetime()


@dataclass(kw_only=True, slots=True)
class ParsePlainDateTimeError(Exception):
    datetime: str

    @override
    def __str__(self) -> str:
        return f"Unable to parse plain datetime; got {self.datetime!r}"


##


def parse_time(time: str, /) -> dt.time:
    """Parse a string into a time."""
    try:
        w_time = Time.parse_common_iso(time)
    except ValueError:
        raise ParseTimeError(time=time) from None
    return w_time.py_time()


@dataclass(kw_only=True, slots=True)
class ParseTimeError(Exception):
    time: str

    @override
    def __str__(self) -> str:
        return f"Unable to parse time; got {self.time!r}"


##


def parse_timedelta(timedelta: str, /) -> dt.timedelta:
    """Parse a string into a timedelta."""
    with suppress(ExtractGroupError):
        rest = extract_group(r"^-([\w\.]+)$", timedelta)
        return -parse_timedelta(rest)
    try:
        days_str, time_str = extract_groups(r"^P(?:(\d+)D)?(?:T([\w\.]*))?$", timedelta)
    except ExtractGroupsError:
        raise _ParseTimedeltaParseError(timedelta=timedelta) from None
    days = ZERO_TIME if days_str == "" else dt.timedelta(days=int(days_str))
    if time_str == "":
        time = ZERO_TIME
    else:
        time_part = DateTimeDelta.parse_common_iso(f"PT{time_str}").time_part()
        _, nanoseconds = divmod(time_part.in_nanoseconds(), 1000)
        if nanoseconds != 0:
            raise _ParseTimedeltaNanosecondError(
                timedelta=timedelta, nanoseconds=nanoseconds
            )
        time = dt.timedelta(microseconds=int(time_part.in_microseconds()))
    return days + time


@dataclass(kw_only=True, slots=True)
class ParseTimedeltaError(Exception):
    timedelta: str


@dataclass(kw_only=True, slots=True)
class _ParseTimedeltaParseError(ParseTimedeltaError):
    @override
    def __str__(self) -> str:
        return f"Unable to parse timedelta; got {self.timedelta!r}"


@dataclass(kw_only=True, slots=True)
class _ParseTimedeltaNanosecondError(ParseTimedeltaError):
    nanoseconds: int

    @override
    def __str__(self) -> str:
        return f"Unable to parse timedelta; got {self.nanoseconds} nanoseconds"


##


def parse_zoned_datetime(datetime: str, /) -> dt.datetime:
    """Parse a string into a zoned datetime."""
    try:
        zdt = ZonedDateTime.parse_common_iso(datetime)
    except ValueError:
        raise ParseZonedDateTimeError(datetime=datetime) from None
    return zdt.py_datetime()


@dataclass(kw_only=True, slots=True)
class ParseZonedDateTimeError(Exception):
    datetime: str

    @override
    def __str__(self) -> str:
        return f"Unable to parse zoned datetime; got {self.datetime!r}"


##


def serialize_date(date: dt.date, /) -> str:
    """Serialize a date."""
    check_date_not_datetime(date)
    return Date.from_py_date(date).format_common_iso()


##


def serialize_datetime(datetime: dt.datetime, /) -> str:
    """Serialize a datetime."""
    try:
        return serialize_plain_datetime(datetime)
    except SerializePlainDateTimeError:
        return serialize_zoned_datetime(datetime)


##


def serialize_duration(duration: Duration, /) -> str:
    """Serialize a duration."""
    if isinstance(duration, int | float):
        return str(duration)
    try:
        return serialize_timedelta(duration)
    except SerializeTimeDeltaError as error:
        raise SerializeDurationError(duration=error.timedelta) from None


@dataclass(kw_only=True, slots=True)
class SerializeDurationError(Exception):
    duration: Duration

    @override
    def __str__(self) -> str:
        return f"Unable to serialize duration; got {self.duration}"


##


def serialize_plain_datetime(datetime: dt.datetime, /) -> str:
    """Serialize a plain datetime."""
    try:
        pdt = PlainDateTime.from_py_datetime(datetime)
    except ValueError:
        raise SerializePlainDateTimeError(datetime=datetime) from None
    return pdt.format_common_iso()


@dataclass(kw_only=True, slots=True)
class SerializePlainDateTimeError(Exception):
    datetime: dt.datetime

    @override
    def __str__(self) -> str:
        return f"Unable to serialize plain datetime; got {self.datetime}"


##


def serialize_time(time: dt.time, /) -> str:
    """Serialize a time."""
    return Time.from_py_time(time).format_common_iso()


##


def serialize_timedelta(timedelta: dt.timedelta, /) -> str:
    """Serialize a timedelta."""
    try:
        dtd = _to_datetime_delta(timedelta)
    except _ToDateTimeDeltaError as error:
        raise SerializeTimeDeltaError(timedelta=error.timedelta) from None
    return dtd.format_common_iso()


@dataclass(kw_only=True, slots=True)
class SerializeTimeDeltaError(Exception):
    timedelta: dt.timedelta

    @override
    def __str__(self) -> str:
        return f"Unable to serialize timedelta; got {self.timedelta}"


##


def serialize_zoned_datetime(datetime: dt.datetime, /) -> str:
    """Serialize a zoned datetime."""
    if datetime.tzinfo is dt.UTC:
        return serialize_zoned_datetime(  # skipif-ci-and-windows
            datetime.replace(tzinfo=UTC)
        )
    try:
        zdt = ZonedDateTime.from_py_datetime(datetime)
    except ValueError:
        raise SerializeZonedDateTimeError(datetime=datetime) from None
    return zdt.format_common_iso()


@dataclass(kw_only=True, slots=True)
class SerializeZonedDateTimeError(Exception):
    datetime: dt.datetime

    @override
    def __str__(self) -> str:
        return f"Unable to serialize zoned datetime; got {self.datetime}"


##


class WheneverLogRecord(LogRecord):
    """Log record powered by `whenever`."""

    zoned_datetime: str

    @override
    def __init__(
        self,
        name: str,
        level: int,
        pathname: str,
        lineno: int,
        msg: object,
        args: Any,
        exc_info: Any,
        func: str | None = None,
        sinfo: str | None = None,
    ) -> None:
        super().__init__(
            name, level, pathname, lineno, msg, args, exc_info, func, sinfo
        )
        length = self._get_length()
        plain = format(self._get_now().to_plain().format_common_iso(), f"{length}s")
        time_zone = self._get_time_zone_key()
        self.zoned_datetime = f"{plain}[{time_zone}]"

    @classmethod
    @cache
    def _get_time_zone(cls) -> ZoneInfo:
        """Get the local timezone."""
        try:
            from utilities.tzlocal import get_local_time_zone
        except ModuleNotFoundError:  # pragma: no cover
            return UTC
        return get_local_time_zone()

    @classmethod
    @cache
    def _get_time_zone_key(cls) -> str:
        """Get the local timezone as a string."""
        return cls._get_time_zone().key

    @classmethod
    @cache
    def _get_length(cls) -> int:
        """Get maximum length of a formatted string."""
        now = cls._get_now().replace(nanosecond=1000).to_plain()
        return len(now.format_common_iso())

    @classmethod
    def _get_now(cls) -> ZonedDateTime:
        """Get the current zoned datetime."""
        return ZonedDateTime.now(cls._get_time_zone().key)


##


def _to_datetime_delta(timedelta: dt.timedelta, /) -> DateTimeDelta:
    """Serialize a timedelta."""
    total_microseconds = datetime_duration_to_microseconds(timedelta)
    if total_microseconds == 0:
        return DateTimeDelta()
    if total_microseconds >= 1:
        days, remainder = divmod(total_microseconds, _MICROSECONDS_PER_DAY)
        seconds, microseconds = divmod(remainder, _MICROSECONDS_PER_SECOND)
        try:
            dtd = DateTimeDelta(days=days, seconds=seconds, microseconds=microseconds)
        except (OverflowError, ValueError):
            raise _ToDateTimeDeltaError(timedelta=timedelta) from None
        months, days, seconds, nanoseconds = dtd.in_months_days_secs_nanos()
        return DateTimeDelta(
            months=months, days=days, seconds=seconds, nanoseconds=nanoseconds
        )
    return -_to_datetime_delta(-timedelta)


@dataclass(kw_only=True, slots=True)
class _ToDateTimeDeltaError(Exception):
    timedelta: dt.timedelta

    @override
    def __str__(self) -> str:
        return f"Unable to create DateTimeDelta; got {self.timedelta}"


__all__ = [
    "MAX_SERIALIZABLE_TIMEDELTA",
    "MIN_SERIALIZABLE_TIMEDELTA",
    "CheckValidZonedDateTimeError",
    "EnsureDateError",
    "EnsureDateTimeError",
    "EnsurePlainDateTimeError",
    "EnsureTimeError",
    "EnsureTimedeltaError",
    "EnsureZonedDateTimeError",
    "ParseDateError",
    "ParseDateTimeError",
    "ParseDurationError",
    "ParsePlainDateTimeError",
    "ParseTimeError",
    "ParseTimedeltaError",
    "ParseZonedDateTimeError",
    "SerializeDurationError",
    "SerializePlainDateTimeError",
    "SerializeTimeDeltaError",
    "SerializeZonedDateTimeError",
    "WheneverLogRecord",
    "check_valid_zoned_datetime",
    "ensure_date",
    "ensure_datetime",
    "ensure_duration",
    "ensure_plain_datetime",
    "ensure_time",
    "ensure_timedelta",
    "ensure_zoned_datetime",
    "parse_date",
    "parse_datetime",
    "parse_duration",
    "parse_plain_datetime",
    "parse_time",
    "parse_timedelta",
    "parse_zoned_datetime",
    "serialize_date",
    "serialize_datetime",
    "serialize_duration",
    "serialize_plain_datetime",
    "serialize_time",
    "serialize_timedelta",
    "serialize_zoned_datetime",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar, dataclass_transform, overload

from pyrsistent import PRecord as _PRecord
from pyrsistent import field as _field
from pyrsistent._field_common import (
    PFIELD_NO_FACTORY,
    PFIELD_NO_INITIAL,
    PFIELD_NO_INVARIANT,
    PFIELD_NO_SERIALIZER,
    PFIELD_NO_TYPE,
)

if TYPE_CHECKING:
    from collections.abc import Callable

    from utilities.types import TypeLike


_T = TypeVar("_T")
_U = TypeVar("_U")


@overload
def field(
    *,
    type: type[_T],
    invariant: Callable[[Any], tuple[bool, Any]] = ...,
    default: Any = ...,
    mandatory: bool = ...,
    factory: Callable[[_U], _U] = ...,
    serializer: Callable[[Any, Any], Any] = ...,
) -> _T: ...
@overload
def field(
    *,
    type: tuple[type[_T]],
    invariant: Callable[[Any], tuple[bool, Any]] = ...,
    default: Any = ...,
    mandatory: bool = ...,
    factory: Callable[[_U], _U] = ...,
    serializer: Callable[[Any, Any], Any] = ...,
) -> _T: ...
@overload
def field(
    *,
    type: tuple[type[_T], type[_U]],
    invariant: Callable[[Any], tuple[bool, Any]] = ...,
    default: Any = ...,
    mandatory: bool = ...,
    factory: Callable[[_U], _U] = ...,
    serializer: Callable[[Any, Any], Any] = ...,
) -> _T | _U: ...
@overload
def field(
    *,
    type: tuple[Any, ...] = ...,
    invariant: Callable[[Any], tuple[bool, Any]] = ...,
    default: Any = ...,
    mandatory: bool = ...,
    factory: Callable[[_U], _U] = ...,
    serializer: Callable[[Any, Any], Any] = ...,
) -> Any: ...
def field(
    *,
    type: TypeLike[_T] = PFIELD_NO_TYPE,  # noqa: A002
    invariant: Callable[[Any], tuple[bool, Any]] = PFIELD_NO_INVARIANT,
    default: Any = PFIELD_NO_INITIAL,
    mandatory: bool = False,
    factory: Callable[[_U], _U] = PFIELD_NO_FACTORY,
    serializer: Callable[[Any, Any], Any] = PFIELD_NO_SERIALIZER,
) -> Any:
    """Field specification factory for :py:class:`PRecord`."""
    return _field(
        type=type,
        invariant=invariant,
        initial=default,
        mandatory=mandatory,
        factory=factory,
        serializer=serializer,
    )


@dataclass_transform(kw_only_default=True, field_specifiers=(field,))
class PRecord(_PRecord): ...


__all__ = ["PRecord", "field"]



from __future__ import annotations

from time import sleep

from hypothesis import example, given
from hypothesis.strategies import floats, integers, none

from utilities.cachetools import TTLSet, cache


class TestCache:
    @example(max_size=None, max_duration=None)
    @example(max_size=None, max_duration=1.0)
    @example(max_size=1, max_duration=None)
    @example(max_size=1, max_duration=1.0)
    @given(max_size=integers(1, 10) | none(), max_duration=floats(0.1, 10.0) | none())
    def test_main(self, *, max_size: int, max_duration: float) -> None:
        counter = 0

        @cache(max_size=max_size, max_duration=max_duration)
        def func(x: int, /) -> int:
            nonlocal counter
            counter += 1
            return x

        for _ in range(2):
            assert func(0) == 0


class TestTTLSet:
    def test_contains(self) -> None:
        set_ = TTLSet(range(3))
        assert 0 in set_
        assert 3 not in set_

    def test_discard(self) -> None:
        set_ = TTLSet(range(3))
        set_.discard(1)
        assert set_ == {0, 2}

    def test_init_with_iterable(self) -> None:
        _ = TTLSet(range(3))

    def test_init_without_iterable(self) -> None:
        _ = TTLSet()

    def test_len(self) -> None:
        set_ = TTLSet(range(3))
        assert len(set_) == 3

    def test_max_duration(self) -> None:
        set_ = TTLSet(range(3), max_duration=0.1)
        assert set_ == {0, 1, 2}
        sleep(0.2)
        assert set_ == set()

    def test_max_size(self) -> None:
        set_ = TTLSet(max_size=3)
        set_.add(1)
        assert set_ == {1}
        set_.add(2)
        assert set_ == {1, 2}
        set_.add(3)
        assert set_ == {1, 2, 3}
        set_.add(4)
        assert set_ == {2, 3, 4}
        set_.add(5)
        assert set_ == {3, 4, 5}

    def test_repr_and_str(self) -> None:
        set_ = TTLSet([None])
        expected = "{None}"
        assert repr(set_) == expected
        assert str(set_) == expected



from __future__ import annotations

from typing import TYPE_CHECKING

from hypothesis import given
from hypothesis.strategies import integers

from utilities.hypothesis import temp_paths, text_ascii
from utilities.shelve import yield_shelf

if TYPE_CHECKING:
    from pathlib import Path


class TestYieldShelf:
    @given(root=temp_paths(), key=text_ascii(), value=integers())
    def test_main(self, *, root: Path, key: str, value: int) -> None:
        path = root.joinpath("shelf")
        with yield_shelf(path) as shelf:
            shelf[key] = value
        with yield_shelf(path) as shelf:
            result = shelf[key]
        assert result == value



from __future__ import annotations

from dataclasses import dataclass
from http import HTTPStatus
from logging import NOTSET, Handler, LogRecord
from typing import TYPE_CHECKING, Any, Self, override

from slack_sdk.webhook.async_client import AsyncWebhookClient

from utilities.asyncio import Looper, timeout_dur
from utilities.datetime import MINUTE, SECOND, datetime_duration_to_float
from utilities.functools import cache
from utilities.math import safe_round
from utilities.sentinel import Sentinel, sentinel

if TYPE_CHECKING:
    from collections.abc import Callable

    from slack_sdk.webhook import WebhookResponse

    from utilities.types import Coroutine1, Duration


_TIMEOUT: Duration = MINUTE


##


async def _send_adapter(url: str, text: str, /) -> None:
    await send_to_slack(url, text)  # pragma: no cover


@dataclass(init=False, unsafe_hash=True)
class SlackHandlerService(Handler, Looper[str]):
    """Service to send messages to Slack."""

    @override
    def __init__(
        self,
        *,
        url: str,
        auto_start: bool = False,
        empty_upon_exit: bool = True,
        freq: Duration = SECOND,
        backoff: Duration = SECOND,
        logger: str | None = None,
        timeout: Duration | None = None,
        _debug: bool = False,
        level: int = NOTSET,
        sender: Callable[[str, str], Coroutine1[None]] = _send_adapter,
        send_timeout: Duration = SECOND,
    ) -> None:
        Looper.__init__(  # Looper first
            self,
            auto_start=auto_start,
            freq=freq,
            empty_upon_exit=empty_upon_exit,
            backoff=backoff,
            logger=logger,
            timeout=timeout,
            _debug=_debug,
        )
        Looper.__post_init__(self)
        Handler.__init__(self, level=level)  # Handler next
        self.url = url
        self.sender = sender
        self.send_timeout = send_timeout

    @override
    def emit(self, record: LogRecord) -> None:
        fmtted = self.format(record)
        try:
            self.put_right_nowait(fmtted)
        except Exception:  # noqa: BLE001  # pragma: no cover
            self.handleError(record)

    @override
    async def core(self) -> None:
        await super().core()
        if self.empty():
            return
        text = "\n".join(self.get_all_nowait())
        async with timeout_dur(duration=self.send_timeout):
            await self.sender(self.url, text)

    @override
    def replace(
        self,
        *,
        auto_start: bool | Sentinel = sentinel,
        empty_upon_exit: bool | Sentinel = sentinel,
        freq: Duration | Sentinel = sentinel,
        backoff: Duration | Sentinel = sentinel,
        logger: str | None | Sentinel = sentinel,
        timeout: Duration | None | Sentinel = sentinel,
        _debug: bool | Sentinel = sentinel,
        **kwargs: Any,
    ) -> Self:
        """Replace elements of the looper."""
        return super().replace(
            url=self.url,
            auto_start=auto_start,
            empty_upon_exit=empty_upon_exit,
            freq=freq,
            backoff=backoff,
            logger=logger,
            timeout=timeout,
            _debug=_debug,
            **kwargs,
        )


##


async def send_to_slack(
    url: str, text: str, /, *, timeout: Duration = _TIMEOUT
) -> None:
    """Send a message via Slack."""
    client = _get_client(url, timeout=timeout)
    async with timeout_dur(duration=timeout):
        response = await client.send(text=text)
    if response.status_code != HTTPStatus.OK:  # pragma: no cover
        raise SendToSlackError(text=text, response=response)


@dataclass(kw_only=True, slots=True)
class SendToSlackError(Exception):
    text: str
    response: WebhookResponse

    @override
    def __str__(self) -> str:
        code = self.response.status_code  # pragma: no cover
        phrase = HTTPStatus(code).phrase  # pragma: no cover
        return f"Error sending to Slack:\n\n{self.text}\n\n{code}: {phrase}"  # pragma: no cover


@cache
def _get_client(url: str, /, *, timeout: Duration = _TIMEOUT) -> AsyncWebhookClient:
    """Get the Slack client."""
    timeout_use = safe_round(datetime_duration_to_float(timeout))
    return AsyncWebhookClient(url, timeout=timeout_use)


__all__ = ["SendToSlackError", "SlackHandlerService", "send_to_slack"]



from __future__ import annotations

from asyncio import sleep
from typing import ClassVar

from utilities.aiolimiter import get_async_limiter
from utilities.text import unique_str
from utilities.timer import Timer


class TestGetAsyncLimiter:
    async def test_main(self) -> None:
        counter = 0

        async def increment() -> None:
            nonlocal counter
            counter += 1
            await sleep(0.01)

        name = unique_str()
        with Timer() as timer:
            for _ in range(2):
                async with get_async_limiter(name, rate=0.5):
                    await increment()
        assert timer >= 0.48

    shared: ClassVar[str] = unique_str()

    async def test_shared1(self) -> None:
        async with get_async_limiter(self.shared):
            await sleep(0.01)

    async def test_shared2(self) -> None:
        async with get_async_limiter(self.shared):
            await sleep(0.01)

    async def test_shared3(self) -> None:
        async with get_async_limiter(self.shared):
            await sleep(0.01)



from __future__ import annotations

from dataclasses import dataclass, field
from functools import total_ordering
from pathlib import Path
from types import NoneType
from typing import Any, Literal, Self, cast, override

from hypothesis import given
from hypothesis.strategies import (
    DataObject,
    booleans,
    data,
    integers,
    lists,
    permutations,
    sampled_from,
)
from polars import DataFrame
from pytest import raises

from tests.test_typing_funcs.no_future import (
    DataClassNoFutureInt,
    DataClassNoFutureIntDefault,
)
from tests.test_typing_funcs.with_future import (
    DataClassFutureInt,
    DataClassFutureIntDefault,
    DataClassFutureIntLowerAndUpper,
    DataClassFutureIntOneAndTwo,
    DataClassFutureListInts,
    DataClassFutureListIntsDefault,
    DataClassFutureLiteral,
    DataClassFutureLiteralNullable,
    DataClassFutureNestedOuterFirstInner,
    DataClassFutureNestedOuterFirstOuter,
    DataClassFutureNone,
    DataClassFutureNoneDefault,
    DataClassFutureTypeLiteral,
    DataClassFutureTypeLiteralNullable,
    TrueOrFalseFutureLit,
    TrueOrFalseFutureTypeLit,
)
from utilities.dataclasses import (
    YieldFieldsError,
    _MappingToDataClassEmptyError,
    _MappingToDataClassMissingValuesError,
    _MappingToDataClassNonUniqueError,
    _OneFieldEmptyError,
    _OneFieldNonUniqueError,
    _parse_dataclass_split_key_value_pairs,
    _ParseDataClassMissingValuesError,
    _ParseDataClassSplitKeyValuePairsDuplicateKeysError,
    _ParseDataClassSplitKeyValuePairsSplitError,
    _ParseDataClassStrMappingToFieldMappingEmptyError,
    _ParseDataClassStrMappingToFieldMappingNonUniqueError,
    _ParseDataClassTextExtraNonUniqueError,
    _ParseDataClassTextParseError,
    _StrMappingToFieldMappingEmptyError,
    _StrMappingToFieldMappingNonUniqueError,
    _YieldFieldsClass,
    _YieldFieldsInstance,
    dataclass_repr,
    dataclass_to_dict,
    is_nullable_lt,
    mapping_to_dataclass,
    one_field,
    parse_dataclass,
    replace_non_sentinel,
    serialize_dataclass,
    str_mapping_to_field_mapping,
    yield_fields,
)
from utilities.functions import get_class_name
from utilities.iterables import one
from utilities.orjson import OrjsonLogRecord
from utilities.polars import are_frames_equal
from utilities.sentinel import sentinel
from utilities.types import Dataclass, StrMapping
from utilities.typing import get_args, is_list_type, is_literal_type, is_optional_type


class TestDataClassRepr:
    def test_overriding_repr(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example(DataClassFutureIntDefault):
            @override
            def __repr__(self) -> str:
                return dataclass_repr(self)

        obj = Example()
        result = repr(obj)
        expected = "Example()"
        assert result == expected

    def test_overriding_repr_defaults(self) -> None:
        @dataclass(kw_only=True)
        class Example(DataClassFutureIntDefault):
            @override
            def __repr__(self) -> str:
                return dataclass_repr(self, defaults=True)

        obj = Example()
        result = repr(obj)
        expected = "Example(int_=0)"
        assert result == expected

    @given(x=integers())
    def test_non_repr_field(self, *, x: int) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: int = field(default=0, repr=False)

        obj = Example(x=x)
        result = dataclass_repr(obj)
        expected = "Example()"
        assert result == expected


class TestDataClassToDictAndDataClassRepr:
    @given(x=integers(), defaults=booleans())
    def test_field_without_defaults(self, *, x: int, defaults: bool) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: int

        obj = Example(x=x)
        dict_res = dataclass_to_dict(obj, defaults=defaults)
        dict_exp = {"x": x}
        assert dict_res == dict_exp
        repr_res = dataclass_repr(obj, defaults=defaults)
        repr_exp = f"Example(x={x})"
        assert repr_res == repr_exp

    @given(x=integers())
    def test_field_with_default_included(self, *, x: int) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: int = 0

        obj = Example(x=x)
        dict_res = dataclass_to_dict(obj, defaults=True)
        dict_exp = {"x": x}
        assert dict_res == dict_exp
        repr_res = dataclass_repr(obj, defaults=True)
        repr_exp = f"Example(x={x})"
        assert repr_res == repr_exp

    def test_field_with_default_dropped(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: int = 0

        obj = Example()
        dict_res = dataclass_to_dict(obj)
        dict_exp = {}
        assert dict_res == dict_exp
        repr_res = dataclass_repr(obj)
        repr_exp = "Example()"
        assert repr_res == repr_exp

    def test_field_with_dataframe_included(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: DataFrame = field(default_factory=DataFrame)

        obj = Example()
        extra = {DataFrame: are_frames_equal}
        dict_res = dataclass_to_dict(
            obj, globalns=globals(), extra=extra, defaults=True
        )
        dict_exp = {"x": DataFrame()}
        assert set(dict_res) == set(dict_exp)
        repr_res = dataclass_repr(obj, globalns=globals(), extra=extra, defaults=True)
        repr_exp = f"Example(x={DataFrame()})"
        assert repr_res == repr_exp

    def test_field_with_dataframe_dropped(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: DataFrame = field(default_factory=DataFrame)

        obj = Example()
        extra = {DataFrame: are_frames_equal}
        dict_res = dataclass_to_dict(obj, globalns=globals(), extra=extra)
        dict_exp = {}
        assert set(dict_res) == set(dict_exp)
        repr_res = dataclass_repr(obj, globalns=globals(), extra=extra)
        repr_exp = "Example()"
        assert repr_res == repr_exp

    @given(x=integers())
    def test_final(self, *, x: int) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            x: int

        def final(obj: type[Dataclass], mapping: StrMapping) -> StrMapping:
            return {f"[{get_class_name(obj)}]": mapping}

        obj = Example(x=x)
        result = dataclass_to_dict(obj, final=final)
        expected = {"[Example]": {"x": x}}
        assert result == expected

    @given(y=integers())
    def test_nested_with_recursive(self, *, y: int) -> None:
        @dataclass(kw_only=True, slots=True)
        class Inner:
            x: int = 0

        @dataclass(kw_only=True, slots=True)
        class Outer:
            inner: Inner
            y: int

        obj = Outer(inner=Inner(), y=y)
        dict_res = dataclass_to_dict(obj, localns=locals(), recursive=True)
        dict_exp = {"inner": {}, "y": y}
        assert dict_res == dict_exp
        repr_res = dataclass_repr(obj, localns=locals(), recursive=True)
        repr_exp = f"Outer(inner=Inner(), y={y})"
        assert repr_res == repr_exp

    @given(y=integers())
    def test_nested_without_recursive(self, *, y: int) -> None:
        @dataclass(kw_only=True, slots=True)
        class Inner:
            x: int = 0

        @dataclass(kw_only=True, slots=True)
        class Outer:
            inner: Inner
            y: int

        obj = Outer(inner=Inner(), y=y)
        dict_res = dataclass_to_dict(obj, localns=locals())
        dict_exp = {"inner": Inner(), "y": y}
        assert dict_res == dict_exp
        repr_res = dataclass_repr(obj, localns=locals())
        repr_exp = f"Outer(inner=TestDataClassToDictAndDataClassRepr.test_nested_without_recursive.<locals>.Inner(x=0), y={y})"
        assert repr_res == repr_exp

    @given(y=lists(integers()), z=integers())
    def test_nested_in_list_with_recursive(self, *, y: list[int], z: int) -> None:
        @dataclass(kw_only=True, slots=True)
        class Inner:
            x: int = 0

        @dataclass(kw_only=True, slots=True)
        class Outer:
            inner: list[Inner]
            y: list[int]
            z: int

        obj = Outer(inner=[Inner()], y=y, z=z)
        dict_res = dataclass_to_dict(obj, localns=locals(), recursive=True)
        dict_exp = {"inner": [{}], "y": y, "z": z}
        assert dict_res == dict_exp
        repr_res = dataclass_repr(obj, localns=locals(), recursive=True)
        repr_exp = f"Outer(inner=[Inner()], y={y}, z={z})"
        assert repr_res == repr_exp

    @given(y=lists(integers()), z=integers())
    def test_nested_in_list_without_recursive(self, *, y: list[int], z: int) -> None:
        @dataclass(kw_only=True, slots=True)
        class Inner:
            x: int = 0

        @dataclass(kw_only=True, slots=True)
        class Outer:
            inner: list[Inner]
            y: list[int]
            z: int

        obj = Outer(inner=[Inner()], y=y, z=z)
        dict_res = dataclass_to_dict(obj, localns=locals())
        dict_exp = {"inner": [Inner(x=0)], "y": y, "z": z}
        assert dict_res == dict_exp
        repr_res = dataclass_repr(obj, localns=locals())
        repr_exp = f"Outer(inner=[TestDataClassToDictAndDataClassRepr.test_nested_in_list_without_recursive.<locals>.Inner(x=0)], y={y}, z={z})"
        assert repr_res == repr_exp


class TestIsNullableLT:
    @given(data=data())
    def test_main(self, *, data: DataObject) -> None:
        @dataclass(kw_only=True)
        @total_ordering
        class Example:
            x: int | None = None

            def __lt__(self, other: Self) -> bool:
                if (cmp := is_nullable_lt(self.x, other.x)) is not None:
                    return cmp
                return False

        obj_none, obj1, obj2 = [Example(x=x) for x in [None, 1, 2]]
        expected = [obj_none, obj_none, obj1, obj1, obj2, obj2]
        result = sorted(data.draw(permutations(expected)))
        assert result == expected


class TestMappingToDataClass:
    @given(key=sampled_from(["int_", "INT_"]), int_=integers())
    def test_exact_match_case_insensitive(self, *, key: str, int_: int) -> None:
        obj = mapping_to_dataclass(DataClassFutureInt, {key: int_})
        expected = DataClassFutureInt(int_=int_)
        assert obj == expected

    @given(key=sampled_from(["in", "IN"]), int_=integers())
    def test_head_case_insensitive(self, *, key: str, int_: int) -> None:
        obj = mapping_to_dataclass(DataClassFutureInt, {key: int_}, head=True)
        expected = DataClassFutureInt(int_=int_)
        assert obj == expected

    @given(int_=integers())
    def test_exact_match_case_sensitive(self, *, int_: int) -> None:
        obj = mapping_to_dataclass(
            DataClassFutureInt, {"int_": int_}, case_sensitive=True
        )
        expected = DataClassFutureInt(int_=int_)
        assert obj == expected

    @given(int_=integers())
    def test_head_case_sensitive(self, *, int_: int) -> None:
        obj = mapping_to_dataclass(
            DataClassFutureInt, {"int": int_}, head=True, case_sensitive=True
        )
        expected = DataClassFutureInt(int_=int_)
        assert obj == expected

    @given(int_=integers())
    def test_extra_key(self, *, int_: int) -> None:
        obj = mapping_to_dataclass(
            DataClassFutureInt, {"int_": int_, "extra": int_}, allow_extra=True
        )
        expected = DataClassFutureInt(int_=int_)
        assert obj == expected

    @given(int_=integers())
    def test_error_exact_match_case_insensitive_empty(self, *, int_: int) -> None:
        with raises(
            _MappingToDataClassEmptyError,
            match=r"Dataclass 'DataClassFutureInt' does not contain a field 'invalid' \(modulo case\)",
        ):
            _ = mapping_to_dataclass(
                DataClassFutureInt, {"int_": int_, "invalid": int_}
            )

    @given(int_=integers())
    def test_error_exact_match_case_insensitive_non_unique(self, *, int_: int) -> None:
        with raises(
            _MappingToDataClassNonUniqueError,
            match=r"Dataclass 'DataClassFutureIntLowerAndUpper' must contain field 'int_' exactly once \(modulo case\); got 'int_', 'INT_' and perhaps more",
        ):
            _ = mapping_to_dataclass(DataClassFutureIntLowerAndUpper, {"int_": int_})

    @given(int_=integers())
    def test_error_head_case_insensitive_empty(self, *, int_: int) -> None:
        with raises(
            _MappingToDataClassEmptyError,
            match=r"Dataclass 'DataClassFutureInt' does not contain any field starting with 'invalid' \(modulo case\)",
        ):
            _ = mapping_to_dataclass(DataClassFutureInt, {"invalid": int_}, head=True)

    @given(int_=integers())
    def test_error_head_case_insensitive_non_unique(self, *, int_: int) -> None:
        with raises(
            _MappingToDataClassNonUniqueError,
            match=r"Dataclass 'DataClassFutureIntOneAndTwo' must contain exactly one field starting with 'int' \(modulo case\); got 'int1', 'int2' and perhaps more",
        ):
            _ = mapping_to_dataclass(
                DataClassFutureIntOneAndTwo, {"int": int_}, head=True
            )

    @given(int_=integers())
    def test_error_exact_match_case_sensitive_empty(self, *, int_: int) -> None:
        with raises(
            _MappingToDataClassEmptyError,
            match=r"Dataclass 'DataClassFutureInt' does not contain a field 'extra'",
        ):
            _ = mapping_to_dataclass(
                DataClassFutureInt, {"int_": int_, "extra": int_}, case_sensitive=True
            )

    # there is no head=False, case_sensitive=True, non-unique case

    @given(int_=integers())
    def test_error_head_case_sensitive_empty(self, *, int_: int) -> None:
        with raises(
            _MappingToDataClassEmptyError,
            match=r"Dataclass 'DataClassFutureInt' does not contain any field starting with 'invalid'",
        ):
            _ = mapping_to_dataclass(
                DataClassFutureInt, {"invalid": int_}, head=True, case_sensitive=True
            )

    @given(int_=integers())
    def test_error_head_case_sensitive_non_unique(self, *, int_: int) -> None:
        with raises(
            _MappingToDataClassNonUniqueError,
            match=r"Dataclass 'DataClassFutureIntOneAndTwo' must contain exactly one field starting with 'int'; got 'int1', 'int2' and perhaps more",
        ):
            _ = mapping_to_dataclass(
                DataClassFutureIntOneAndTwo,
                {"int": int_},
                head=True,
                case_sensitive=True,
            )

    def test_error_missing_values(self) -> None:
        with raises(
            _MappingToDataClassMissingValuesError,
            match="Unable to construct 'DataClassFutureInt'; missing values for 'int_'",
        ):
            _ = mapping_to_dataclass(DataClassFutureInt, {})


class TestOneField:
    @given(key=sampled_from(["int_", "INT_"]))
    def test_exact_match_case_insensitive(self, *, key: str) -> None:
        obj = one_field(DataClassFutureInt, key)
        expected = one(yield_fields(DataClassFutureInt))
        assert obj == expected

    @given(key=sampled_from(["in", "IN"]))
    def test_head_case_insensitive(self, *, key: str) -> None:
        obj = one_field(DataClassFutureInt, key, head=True)
        expected = one(yield_fields(DataClassFutureInt))
        assert obj == expected

    def test_exact_match_case_sensitive(self) -> None:
        obj = one_field(DataClassFutureInt, "int_", case_sensitive=True)
        expected = one(yield_fields(DataClassFutureInt))
        assert obj == expected

    def test_head_case_sensitive(self) -> None:
        obj = one_field(DataClassFutureInt, "int", head=True, case_sensitive=True)
        expected = one(yield_fields(DataClassFutureInt))
        assert obj == expected

    def test_error_exact_match_case_insensitive_empty(self) -> None:
        with raises(
            _OneFieldEmptyError,
            match=r"Dataclass 'DataClassFutureInt' does not contain a field 'invalid' \(modulo case\)",
        ):
            _ = one_field(DataClassFutureInt, "invalid")

    def test_error_exact_match_case_insensitive_non_unique(self) -> None:
        with raises(
            _OneFieldNonUniqueError,
            match=r"Dataclass 'DataClassFutureIntLowerAndUpper' must contain field 'int_' exactly once \(modulo case\); got 'int_', 'INT_' and perhaps more",
        ):
            _ = one_field(DataClassFutureIntLowerAndUpper, "int_")

    def test_error_head_case_insensitive_empty(self) -> None:
        with raises(
            _OneFieldEmptyError,
            match=r"Dataclass 'DataClassFutureInt' does not contain any field starting with 'invalid' \(modulo case\)",
        ):
            _ = one_field(DataClassFutureInt, "invalid", head=True)

    def test_error_head_case_insensitive_non_unique(self) -> None:
        with raises(
            _OneFieldNonUniqueError,
            match=r"Dataclass 'DataClassFutureIntOneAndTwo' must contain exactly one field starting with 'int' \(modulo case\); got 'int1', 'int2' and perhaps more",
        ):
            _ = one_field(DataClassFutureIntOneAndTwo, "int", head=True)

    def test_error_exact_match_case_sensitive_empty(self) -> None:
        with raises(
            _OneFieldEmptyError,
            match=r"Dataclass 'DataClassFutureInt' does not contain a field 'INT_'",
        ):
            _ = one_field(DataClassFutureInt, "INT_", case_sensitive=True)

    # there is no head=False, case_sensitive=True, non-unique case

    def test_error_head_case_sensitive_empty(self) -> None:
        with raises(
            _OneFieldEmptyError,
            match=r"Dataclass 'DataClassFutureInt' does not contain any field starting with 'INT_'",
        ):
            _ = one_field(DataClassFutureInt, "INT_", head=True, case_sensitive=True)

    def test_error_head_case_sensitive_non_unique(self) -> None:
        with raises(
            _OneFieldNonUniqueError,
            match=r"Dataclass 'DataClassFutureIntOneAndTwo' must contain exactly one field starting with 'int'; got 'int1', 'int2' and perhaps more",
        ):
            _ = one_field(
                DataClassFutureIntOneAndTwo, "int", head=True, case_sensitive=True
            )


class TestParseDataClassSplitKeyValuePairs:
    @given(text=sampled_from(["a=1,b=22,c=333", "{a=1,b=22,c=333}"]))
    def test_main(self, *, text: str) -> None:
        result = _parse_dataclass_split_key_value_pairs(text, DataClassFutureInt)
        expected = {"a": "1", "b": "22", "c": "333"}
        assert result == expected


class TestReplaceNonSentinel:
    def test_main(self) -> None:
        obj = DataClassFutureIntDefault()
        assert obj.int_ == 0
        obj1 = replace_non_sentinel(obj, int_=1)
        assert obj1.int_ == 1
        obj2 = replace_non_sentinel(obj1, int_=sentinel)
        assert obj2.int_ == 1

    def test_in_place(self) -> None:
        obj = DataClassFutureIntDefault()
        assert obj.int_ == 0
        replace_non_sentinel(obj, int_=1, in_place=True)
        assert obj.int_ == 1
        replace_non_sentinel(obj, int_=sentinel, in_place=True)
        assert obj.int_ == 1


class TestSerializeAndParseDataClass:
    @given(int_=integers())
    def test_main_future_int(self, *, int_: int) -> None:
        obj = DataClassFutureInt(int_=int_)
        serialized = serialize_dataclass(obj)
        result = parse_dataclass(serialized, DataClassFutureInt)
        assert result == obj

    def test_main_future_int_default(self) -> None:
        obj = DataClassFutureIntDefault()
        serialized = serialize_dataclass(obj)
        result = parse_dataclass(serialized, DataClassFutureIntDefault)
        assert result == obj

    @given(int_=integers())
    def test_literal_type(self, *, int_: int) -> None:
        obj = DataClassFutureInt(int_=int_)
        serialized = serialize_dataclass(obj)
        result = parse_dataclass(
            serialized,
            DataClassFutureInt,
            extra_parsers={Literal["lit"]: NotImplementedError},
        )
        assert result == obj

    @given(int_=integers())
    def test_type_extra(self, *, int_: int) -> None:
        obj = DataClassFutureNestedOuterFirstOuter(
            inner=DataClassFutureNestedOuterFirstInner(int_=int_)
        )

        def serializer(obj: DataClassFutureNestedOuterFirstInner, /) -> str:
            return serialize_dataclass(obj)

        serialized = serialize_dataclass(
            obj, extra_serializers={DataClassFutureNestedOuterFirstInner: serializer}
        )
        result = parse_dataclass(
            serialized,
            DataClassFutureNestedOuterFirstOuter,
            globalns=globals(),
            extra_parsers={
                DataClassFutureNestedOuterFirstInner: lambda text: parse_dataclass(
                    text, DataClassFutureNestedOuterFirstInner
                )
            },
        )
        assert result == obj

    @given(key=sampled_from(["int_", "INT_"]), int_=integers())
    def test_parse_text_case_insensitive(self, *, key: str, int_: int) -> None:
        result = parse_dataclass(f"{key}={int_}", DataClassFutureInt)
        expected = DataClassFutureInt(int_=int_)
        assert result == expected

    @given(int_=integers())
    def test_parse_text_case_sensitive(self, *, int_: int) -> None:
        result = parse_dataclass(
            f"int_={int_}", DataClassFutureInt, case_sensitive=True
        )
        expected = DataClassFutureInt(int_=int_)
        assert result == expected

    @given(key=sampled_from(["int_", "INT_"]), int_=integers())
    def test_parse_mapping_case_insensitive(self, *, key: str, int_: int) -> None:
        result = parse_dataclass({key: str(int_)}, DataClassFutureInt)
        expected = DataClassFutureInt(int_=int_)
        assert result == expected

    @given(int_=integers())
    def test_parse_mapping_case_sensitive(self, *, int_: int) -> None:
        result = parse_dataclass(
            {"int_": str(int_)}, DataClassFutureInt, case_sensitive=True
        )
        expected = DataClassFutureInt(int_=int_)
        assert result == expected

    def test_parser_split_key_value_pairs_split(self) -> None:
        with raises(
            _ParseDataClassSplitKeyValuePairsSplitError,
            match="Unable to construct 'DataClassFutureInt'; failed to split key-value pair 'bbb'",
        ):
            _ = parse_dataclass("a=1,bbb,c=333", DataClassFutureInt)

    def test_error_parse_split_key_value_pairs_duplicate(self) -> None:
        with raises(
            _ParseDataClassSplitKeyValuePairsDuplicateKeysError,
            match=r"Unable to construct 'DataClassFutureInt' since there are duplicate keys; got \{'b': 2\}",
        ):
            _ = parse_dataclass("a=1,b=22a,b=22b,c=3", DataClassFutureInt)

    def test_error_text_parse(self) -> None:
        with raises(
            _ParseDataClassTextParseError,
            match="Unable to construct 'DataClassFutureInt' since the field 'int_' of type <class 'int'> could not be parsed; got 'invalid'",
        ):
            _ = parse_dataclass("int_=invalid", DataClassFutureInt)

    def test_error_text_extra_non_unique(self) -> None:
        with raises(
            _ParseDataClassTextExtraNonUniqueError,
            match="Unable to construct 'DataClassFutureInt' since the field 'int_' of type <class 'int'> must contain exactly one parent class in `extra`; got <class 'int'>, <class 'int'> and perhaps more",
        ):
            _ = parse_dataclass(
                "int_=0",
                DataClassFutureInt,
                extra_parsers={int | str: int, int | float: int},
            )

    @given(int_=integers())
    def test_error_exact_match_case_insensitive_empty(self, *, int_: int) -> None:
        with raises(
            _ParseDataClassStrMappingToFieldMappingEmptyError,
            match=r"Unable to construct 'DataClassFutureInt' since it does not contain a field 'invalid' \(modulo case\)",
        ):
            _ = parse_dataclass(f"int_={int_},invalid={int_}", DataClassFutureInt)

    @given(int_=integers())
    def test_error_exact_match_case_insensitive_non_unique(self, *, int_: int) -> None:
        with raises(
            _ParseDataClassStrMappingToFieldMappingNonUniqueError,
            match=r"Unable to construct 'DataClassFutureIntLowerAndUpper' since it must contain field 'int_' exactly once \(modulo case\); got 'int_', 'INT_' and perhaps more",
        ):
            _ = parse_dataclass(f"int_={int_}", DataClassFutureIntLowerAndUpper)

    @given(int_=integers())
    def test_error_head_case_insensitive_empty(self, *, int_: int) -> None:
        with raises(
            _ParseDataClassStrMappingToFieldMappingEmptyError,
            match=r"Unable to construct 'DataClassFutureInt' since it does not contain any field starting with 'invalid' \(modulo case\)",
        ):
            _ = parse_dataclass(f"invalid={int_}", DataClassFutureInt, head=True)

    @given(int_=integers())
    def test_error_head_case_insensitive_non_unique(self, *, int_: int) -> None:
        with raises(
            _ParseDataClassStrMappingToFieldMappingNonUniqueError,
            match=r"Unable to construct 'DataClassFutureIntOneAndTwo' since it must contain exactly one field starting with 'int' \(modulo case\); got 'int1', 'int2' and perhaps more",
        ):
            _ = parse_dataclass(f"int={int_}", DataClassFutureIntOneAndTwo, head=True)

    @given(int_=integers())
    def test_error_exact_match_case_sensitive_empty(self, *, int_: int) -> None:
        with raises(
            _ParseDataClassStrMappingToFieldMappingEmptyError,
            match=r"Unable to construct 'DataClassFutureInt' since it does not contain a field 'extra'",
        ):
            _ = parse_dataclass(
                f"int_={int_},extra={int_}", DataClassFutureInt, case_sensitive=True
            )

    # there is no head=False, case_sensitive=True, non-unique case

    @given(int_=integers())
    def test_error_head_case_sensitive_empty(self, *, int_: int) -> None:
        with raises(
            _ParseDataClassStrMappingToFieldMappingEmptyError,
            match=r"Unable to construct 'DataClassFutureInt' since it does not contain any field starting with 'invalid'",
        ):
            _ = parse_dataclass(
                f"invalid={int_}", DataClassFutureInt, head=True, case_sensitive=True
            )

    @given(int_=integers())
    def test_error_head_case_sensitive_non_unique(self, *, int_: int) -> None:
        with raises(
            _ParseDataClassStrMappingToFieldMappingNonUniqueError,
            match=r"Unable to construct 'DataClassFutureIntOneAndTwo' since it must contain exactly one field starting with 'int'; got 'int1', 'int2' and perhaps more",
        ):
            _ = parse_dataclass(
                f"int={int_}",
                DataClassFutureIntOneAndTwo,
                head=True,
                case_sensitive=True,
            )

    def test_error_missing_values(self) -> None:
        with raises(
            _ParseDataClassMissingValuesError,
            match="Unable to construct 'DataClassFutureInt'; missing values for 'int_'",
        ):
            _ = parse_dataclass("", DataClassFutureInt)


class TestStrMappingToFieldMapping:
    @given(key=sampled_from(["int_", "INT_"]), int_=integers())
    def test_exact_match_case_insensitive(self, *, key: str, int_: int) -> None:
        result = str_mapping_to_field_mapping(DataClassFutureInt, {key: int_})
        assert len(result) == 1
        assert one(result) == one(yield_fields(DataClassFutureInt))
        assert one(result.values()) == int_

    @given(key=sampled_from(["in", "IN"]), int_=integers())
    def test_head_case_insensitive(self, *, key: str, int_: int) -> None:
        result = str_mapping_to_field_mapping(
            DataClassFutureInt, {key: int_}, head=True
        )
        assert len(result) == 1
        assert one(result) == one(yield_fields(DataClassFutureInt))
        assert one(result.values()) == int_

    @given(int_=integers())
    def test_exact_match_case_sensitive(self, *, int_: int) -> None:
        result = str_mapping_to_field_mapping(
            DataClassFutureInt, {"int_": int_}, case_sensitive=True
        )
        assert len(result) == 1
        assert one(result) == one(yield_fields(DataClassFutureInt))
        assert one(result.values()) == int_

    @given(int_=integers())
    def test_head_case_sensitive(self, *, int_: int) -> None:
        result = str_mapping_to_field_mapping(
            DataClassFutureInt, {"int": int_}, head=True, case_sensitive=True
        )
        assert len(result) == 1
        assert one(result) == one(yield_fields(DataClassFutureInt))
        assert one(result.values()) == int_

    @given(int_=integers())
    def test_extra_key(self, *, int_: int) -> None:
        result = str_mapping_to_field_mapping(
            DataClassFutureInt, {"int_": int_, "extra": int_}, allow_extra=True
        )
        assert len(result) == 1
        assert one(result) == one(yield_fields(DataClassFutureInt))
        assert one(result.values()) == int_

    @given(int_=integers())
    def test_error_exact_match_case_insensitive_empty(self, *, int_: int) -> None:
        with raises(
            _StrMappingToFieldMappingEmptyError,
            match=r"Dataclass 'DataClassFutureInt' does not contain a field 'invalid' \(modulo case\)",
        ):
            _ = str_mapping_to_field_mapping(
                DataClassFutureInt, {"int_": int_, "invalid": int_}
            )

    @given(int_=integers())
    def test_error_exact_match_case_insensitive_non_unique(self, *, int_: int) -> None:
        with raises(
            _StrMappingToFieldMappingNonUniqueError,
            match=r"Dataclass 'DataClassFutureIntLowerAndUpper' must contain field 'int_' exactly once \(modulo case\); got 'int_', 'INT_' and perhaps more",
        ):
            _ = str_mapping_to_field_mapping(
                DataClassFutureIntLowerAndUpper, {"int_": int_}
            )

    @given(int_=integers())
    def test_error_head_case_insensitive_empty(self, *, int_: int) -> None:
        with raises(
            _StrMappingToFieldMappingEmptyError,
            match=r"Dataclass 'DataClassFutureInt' does not contain any field starting with 'invalid' \(modulo case\)",
        ):
            _ = str_mapping_to_field_mapping(
                DataClassFutureInt, {"invalid": int_}, head=True
            )

    @given(int_=integers())
    def test_error_head_case_insensitive_non_unique(self, *, int_: int) -> None:
        with raises(
            _StrMappingToFieldMappingNonUniqueError,
            match=r"Dataclass 'DataClassFutureIntOneAndTwo' must contain exactly one field starting with 'int' \(modulo case\); got 'int1', 'int2' and perhaps more",
        ):
            _ = str_mapping_to_field_mapping(
                DataClassFutureIntOneAndTwo, {"int": int_}, head=True
            )

    @given(int_=integers())
    def test_error_exact_match_case_sensitive_empty(self, *, int_: int) -> None:
        with raises(
            _StrMappingToFieldMappingEmptyError,
            match=r"Dataclass 'DataClassFutureInt' does not contain a field 'extra'",
        ):
            _ = str_mapping_to_field_mapping(
                DataClassFutureInt, {"int_": int_, "extra": int_}, case_sensitive=True
            )

    # there is no head=False, case_sensitive=True, non-unique case

    @given(int_=integers())
    def test_error_head_case_sensitive_empty(self, *, int_: int) -> None:
        with raises(
            _StrMappingToFieldMappingEmptyError,
            match=r"Dataclass 'DataClassFutureInt' does not contain any field starting with 'invalid'",
        ):
            _ = str_mapping_to_field_mapping(
                DataClassFutureInt, {"invalid": int_}, head=True, case_sensitive=True
            )

    @given(int_=integers())
    def test_error_head_case_sensitive_non_unique(self, *, int_: int) -> None:
        with raises(
            _StrMappingToFieldMappingNonUniqueError,
            match=r"Dataclass 'DataClassFutureIntOneAndTwo' must contain exactly one field starting with 'int'; got 'int1', 'int2' and perhaps more",
        ):
            _ = str_mapping_to_field_mapping(
                DataClassFutureIntOneAndTwo,
                {"int": int_},
                head=True,
                case_sensitive=True,
            )


class TestYieldFields:
    def test_class_no_future_int(self) -> None:
        result = one(yield_fields(DataClassNoFutureInt))
        expected = _YieldFieldsClass(name="int_", type_=int, kw_only=True)
        assert result == expected

    def test_class_no_future_int_default(self) -> None:
        result = one(yield_fields(DataClassNoFutureIntDefault))
        expected = _YieldFieldsClass(name="int_", type_=int, default=0, kw_only=True)
        assert result == expected

    def test_class_future_none(self) -> None:
        result = one(yield_fields(DataClassFutureNone))
        expected = _YieldFieldsClass(name="none", type_=NoneType, kw_only=True)
        assert result == expected

    def test_class_future_non_default(self) -> None:
        result = one(yield_fields(DataClassFutureNoneDefault))
        expected = _YieldFieldsClass(
            name="none", type_=NoneType, default=None, kw_only=True
        )
        assert result == expected

    def test_class_future_int(self) -> None:
        result = one(yield_fields(DataClassFutureInt))
        expected = _YieldFieldsClass(name="int_", type_=int, kw_only=True)
        assert result == expected

    def test_class_future_list_ints(self) -> None:
        result = one(yield_fields(DataClassFutureListInts))
        expected = _YieldFieldsClass(name="ints", type_=list[int], kw_only=True)
        assert result == expected
        assert is_list_type(result.type_)
        assert get_args(result.type_) == (int,)

    def test_class_future_list_ints_default(self) -> None:
        result = one(yield_fields(DataClassFutureListIntsDefault))
        expected = _YieldFieldsClass(
            name="ints", type_=list[int], default_factory=list, kw_only=True
        )
        assert result == expected
        assert is_list_type(result.type_)
        assert get_args(result.type_) == (int,)

    def test_class_future_literal(self) -> None:
        result = one(yield_fields(DataClassFutureLiteral, globalns=globals()))
        expected = _YieldFieldsClass(
            name="truth", type_=TrueOrFalseFutureLit, kw_only=True
        )
        assert result == expected
        assert is_literal_type(result.type_)
        assert get_args(result.type_) == ("true", "false")

    def test_class_future_literal_nullable(self) -> None:
        result = one(yield_fields(DataClassFutureLiteralNullable, globalns=globals()))
        expected = _YieldFieldsClass(
            name="truth", type_=TrueOrFalseFutureLit | None, default=None, kw_only=True
        )
        assert result == expected
        assert is_optional_type(result.type_)
        args = get_args(result.type_)
        assert args == (TrueOrFalseFutureLit, NoneType)
        assert get_args(args[0]) == ("true", "false")

    def test_class_future_nested(self) -> None:
        result = one(
            yield_fields(DataClassFutureNestedOuterFirstOuter, globalns=globals())
        )
        expected = _YieldFieldsClass(
            name="inner", type_=DataClassFutureNestedOuterFirstInner, kw_only=True
        )
        assert result == expected
        assert result.type_ is DataClassFutureNestedOuterFirstInner

    def test_class_future_type_literal(self) -> None:
        result = one(
            yield_fields(
                DataClassFutureTypeLiteral, globalns=globals(), localns=locals()
            )
        )
        expected = _YieldFieldsClass(
            name="truth", type_=TrueOrFalseFutureTypeLit, kw_only=True
        )
        assert result == expected
        assert is_literal_type(result.type_)
        assert get_args(result.type_) == ("true", "false")

    def test_class_future_type_literal_nullable(self) -> None:
        result = one(
            yield_fields(DataClassFutureTypeLiteralNullable, globalns=globals())
        )
        expected = _YieldFieldsClass(
            name="truth",
            type_=TrueOrFalseFutureTypeLit | None,
            default=None,
            kw_only=True,
        )
        assert result == expected
        assert is_optional_type(result.type_)
        args = get_args(result.type_)
        assert args == (TrueOrFalseFutureTypeLit, NoneType)
        assert get_args(args[0]) == ("true", "false")

    def test_class_orjson_log_record(self) -> None:
        result = list(yield_fields(OrjsonLogRecord, globalns=globals()))
        exp_head = [
            _YieldFieldsClass(name="name", type_=str, kw_only=True),
            _YieldFieldsClass(name="message", type_=str, kw_only=True),
            _YieldFieldsClass(name="level", type_=int, kw_only=True),
        ]
        assert result[:3] == exp_head
        exp_tail = [
            _YieldFieldsClass(
                name="extra", type_=StrMapping | None, default=None, kw_only=True
            ),
            _YieldFieldsClass(
                name="log_file", type_=Path | None, default=None, kw_only=True
            ),
            _YieldFieldsClass(
                name="log_file_line_num", type_=int | None, default=None, kw_only=True
            ),
        ]
        assert result[-3:] == exp_tail

    @given(int_=integers())
    def test_instance_no_future_int(self, *, int_: int) -> None:
        obj = DataClassNoFutureInt(int_=int_)
        result = one(yield_fields(obj))
        expected = _YieldFieldsInstance(
            name="int_", value=int_, type_=int, kw_only=True
        )
        assert result == expected

    @given(int_=integers())
    def test_instance_no_future_int_default(self, *, int_: int) -> None:
        obj = DataClassNoFutureIntDefault(int_=int_)
        result = one(yield_fields(obj))
        expected = _YieldFieldsInstance(
            name="int_", value=int_, type_=int, default=0, kw_only=True
        )
        assert result == expected

    def test_instance_future_none_default(self) -> None:
        obj = DataClassFutureNoneDefault()
        result = one(yield_fields(obj))
        expected = _YieldFieldsInstance(
            name="none", value=None, type_=NoneType, default=None, kw_only=True
        )
        assert result == expected

    @given(int_=integers())
    def test_instance_future_int(self, *, int_: int) -> None:
        obj = DataClassFutureInt(int_=int_)
        field = one(yield_fields(obj))
        assert not field.equals_default()
        assert field.keep()
        assert not field.keep(include=[])
        assert not field.keep(exclude=["int_"])

    @given(int_=integers())
    def test_instance_with_default_equals_default(self, *, int_: int) -> None:
        obj = DataClassFutureIntDefault(int_=int_)
        field = one(yield_fields(obj))
        result = field.equals_default()
        expected = int_ == 0
        assert result is expected
        assert field.keep() is not expected
        assert field.keep(defaults=True)

    @given(ints=lists(integers()))
    def test_instance_future_list_ints_default(self, *, ints: list[int]) -> None:
        obj = DataClassFutureListIntsDefault(ints=ints)
        field = one(yield_fields(obj))
        result = field.equals_default()
        expected = ints == []
        assert result is expected
        assert field.keep() is not expected

    def test_error(self) -> None:
        with raises(
            YieldFieldsError,
            match="Object must be a dataclass instance or class; got None",
        ):
            _ = list(yield_fields(cast("Any", None)))



from __future__ import annotations

import datetime as dt
from logging import getLogger
from typing import TYPE_CHECKING

from tzlocal import get_localzone

if TYPE_CHECKING:
    from zoneinfo import ZoneInfo


def get_local_time_zone() -> ZoneInfo:
    """Get the local time zone, with the logging disabled."""
    logger = getLogger("tzlocal")  # avoid import cycle
    init_disabled = logger.disabled
    logger.disabled = True
    time_zone = get_localzone()
    logger.disabled = init_disabled
    return time_zone


def get_now_local() -> dt.datetime:
    """Get the current local time."""
    return dt.datetime.now(tz=get_local_time_zone())


NOW_LOCAL = get_now_local()


def get_today_local() -> dt.date:
    """Get the current, timezone-aware local date."""
    return get_now_local().date()


TODAY_LOCAL = get_today_local()


__all__ = [
    "NOW_LOCAL",
    "TODAY_LOCAL",
    "get_local_time_zone",
    "get_now_local",
    "get_today_local",
]



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Literal, override

from pytest import approx

from tests.conftest import IS_CI
from utilities.asyncio import Looper
from utilities.contextlib import suppress_super_object_attribute_error
from utilities.datetime import MILLISECOND

if TYPE_CHECKING:
    from collections.abc import Iterator

    from utilities.types import Duration

_FREQ: Duration = 10 * MILLISECOND
_BACKOFF: Duration = 100 * MILLISECOND
_REL: float = 2.0 if IS_CI else 0.25


# assert


def assert_looper_stats(
    looper: Looper[Any],
    /,
    *,
    entries: int = 0,
    core_successes: int | tuple[Literal[">="], int] = 0,
    core_failures: int = 0,
    initialization_successes: int = 0,
    initialization_failures: int = 0,
    tear_down_successes: int = 0,
    tear_down_failures: int = 0,
    restart_successes: int = 0,
    restart_failures: int = 0,
    stops: int = 0,
    rel: float = _REL,
) -> None:
    stats = looper.stats
    assert stats.entries == entries, f"{stats=}, {entries=}"
    assert stats.core_attempts == (stats.core_successes + stats.core_failures), (
        f"{stats=}"
    )
    match core_successes:
        case int():
            assert stats.core_successes == approx(core_successes, rel=rel), (
                f"{stats=}, {core_successes=}"
            )
        case ">=", int() as min_successes:
            assert stats.core_successes >= min_successes, f"{stats=}, {min_successes=}"
    assert stats.core_failures == approx(core_failures, rel=rel), (
        f"{stats=}, {core_failures=}"
    )
    assert stats.initialization_attempts == (
        stats.initialization_successes + stats.initialization_failures
    ), f"{stats=}"
    assert stats.initialization_successes == approx(
        initialization_successes, rel=rel
    ), f"{stats=}, {initialization_successes=}"
    assert stats.initialization_failures == approx(initialization_failures, rel=rel), (
        f"{stats=}, {initialization_failures=}"
    )
    assert stats.tear_down_attempts == (
        stats.tear_down_successes + stats.tear_down_failures
    ), f"{stats=}"
    assert stats.tear_down_successes == approx(tear_down_successes, rel=rel), (
        f"{stats=}, {tear_down_successes=}"
    )
    assert stats.tear_down_failures == approx(tear_down_failures, rel=rel), (
        f"{stats=}, {initialization_failures=}"
    )
    assert stats.restart_attempts == (
        stats.restart_successes + stats.restart_failures
    ), f"{stats=}"
    assert stats.restart_successes == approx(restart_successes, rel=rel), (
        f"{stats=}, {restart_successes=}"
    )
    assert stats.restart_failures == approx(restart_failures, rel=rel), (
        f"{stats=}, {restart_failures=}"
    )
    assert stats.stops == stops, f"{stats=}, {stops=}"


def assert_looper_full(
    looper: Looper[Any], /, *, stops: int = 0, rel: float = _REL
) -> None:
    assert_looper_stats(
        looper,
        entries=1,
        core_successes=99,
        initialization_successes=1,
        stops=stops,
        rel=rel,
    )


# counting looper


@dataclass(kw_only=True)
class CountingLooper(Looper[Any]):
    freq: Duration = field(default=_FREQ, repr=False)
    backoff: Duration = field(default=_BACKOFF, repr=False)
    _debug: bool = field(default=True, repr=False)
    count: int = 0
    max_count: int = 10

    @override
    async def _initialize_core(self) -> None:
        await super()._initialize_core()
        self.count = 0

    @override
    async def core(self) -> None:
        await super().core()
        self.count += 1
        if self.count >= self.max_count:
            raise CountingLooperError


class CountingLooperError(Exception): ...


# one sub looper


@dataclass(kw_only=True)
class OuterCountingLooper(CountingLooper):
    inner: CountingLooper = field(init=False, repr=False)
    inner_auto_start: bool = False

    @override
    def __post_init__(self) -> None:
        super().__post_init__()
        self.inner = CountingLooper(
            auto_start=self.inner_auto_start,
            freq=self.freq / 2,
            backoff=self.backoff / 2,
            max_count=round(self.max_count / 2),
        )

    @override
    def _yield_sub_loopers(self) -> Iterator[Looper]:
        yield from super()._yield_sub_loopers()
        yield self.inner


# two sub loopers


@dataclass(kw_only=True)
class MultipleSubLoopers(CountingLooper):
    inner1: CountingLooper = field(init=False, repr=False)
    inner2: CountingLooper = field(init=False, repr=False)
    inner1_auto_start: bool = False
    inner2_auto_start: bool = False

    @override
    def __post_init__(self) -> None:
        super().__post_init__()
        self.inner1 = CountingLooper(
            auto_start=self.inner1_auto_start,
            freq=self.freq / 2,
            backoff=self.backoff / 2,
            max_count=round(self.max_count / 2),
        )
        self.inner2 = CountingLooper(
            auto_start=self.inner2_auto_start,
            freq=self.freq / 3,
            backoff=self.backoff / 3,
            max_count=round(self.max_count / 3),
        )

    @override
    def _yield_sub_loopers(self) -> Iterator[Looper]:
        yield from super()._yield_sub_loopers()
        yield self.inner1
        yield self.inner2


# nested sub loopers


@dataclass(kw_only=True)
class Outer2CountingLooper(CountingLooper):
    middle: OuterCountingLooper = field(init=False, repr=False)
    middle_auto_start: bool = False
    inner_auto_start: bool = False

    @override
    def __post_init__(self) -> None:
        super().__post_init__()
        self.middle = OuterCountingLooper(
            auto_start=self.middle_auto_start,
            freq=self.freq / 2,
            backoff=self.backoff / 2,
            max_count=round(self.max_count / 2),
            inner_auto_start=self.inner_auto_start,
        )

    @override
    def _yield_sub_loopers(self) -> Iterator[Looper]:
        yield from super()._yield_sub_loopers()
        yield self.middle


# one mixin


@dataclass(kw_only=True)
class CounterMixin:
    freq: Duration = field(default=_FREQ, repr=False)
    backoff: Duration = field(default=_BACKOFF, repr=False)
    _debug: bool = field(default=True, repr=False)
    count: int = 0
    max_count: int = 10
    counter_auto_start: bool = False
    _counter: CountingLooper = field(init=False, repr=False)

    def __post_init__(self) -> None:
        with suppress_super_object_attribute_error():
            super().__post_init__()  # pyright: ignore[reportAttributeAccessIssue]
        self._counter = CountingLooper(
            auto_start=self.counter_auto_start,
            freq=self.freq / 2,
            backoff=self.backoff / 2,
            max_count=round(self.max_count / 2),
        )

    def _yield_sub_loopers(self) -> Iterator[Looper[Any]]:
        with suppress_super_object_attribute_error():
            yield from super()._yield_sub_loopers()  # pyright: ignore[reportAttributeAccessIssue]
        yield self._counter


@dataclass(kw_only=True)
class LooperWithCounterMixin(CounterMixin, Looper): ...


# two mixins


@dataclass(kw_only=True)
class CounterMixin1:
    freq: Duration = field(default=_FREQ, repr=False)
    backoff: Duration = field(default=_BACKOFF, repr=False)
    _debug: bool = field(default=True, repr=False)
    count: int = 0
    max_count: int = 10
    counter1_auto_start: bool = False
    _counter1: CountingLooper = field(init=False, repr=False)

    def __post_init__(self) -> None:
        with suppress_super_object_attribute_error():
            super().__post_init__()  # pyright: ignore[reportAttributeAccessIssue]
        self._counter1 = CountingLooper(
            auto_start=self.counter1_auto_start,
            freq=self.freq / 2,
            backoff=self.backoff / 2,
            max_count=round(self.max_count / 2),
        )

    def _yield_sub_loopers(self) -> Iterator[Looper[Any]]:
        with suppress_super_object_attribute_error():
            yield from super()._yield_sub_loopers()  # pyright: ignore[reportAttributeAccessIssue]
        yield self._counter1


@dataclass(kw_only=True)
class CounterMixin2:
    freq: Duration = field(default=_FREQ, repr=False)
    backoff: Duration = field(default=_BACKOFF, repr=False)
    _debug: bool = field(default=True, repr=False)
    count: int = 0
    max_count: int = 10
    counter2_auto_start: bool = False
    _counter2: CountingLooper = field(init=False, repr=False)

    def __post_init__(self) -> None:
        with suppress_super_object_attribute_error():
            super().__post_init__()  # pyright: ignore[reportAttributeAccessIssue]
        self._counter2 = CountingLooper(
            auto_start=self.counter2_auto_start,
            freq=self.freq / 3,
            backoff=self.backoff / 3,
            max_count=round(self.max_count / 3),
        )

    def _yield_sub_loopers(self) -> Iterator[Looper[Any]]:
        with suppress_super_object_attribute_error():
            yield from super()._yield_sub_loopers()  # pyright: ignore[reportAttributeAccessIssue]
        yield self._counter2


@dataclass(kw_only=True)
class LooperWithCounterMixins(CounterMixin1, CounterMixin2, Looper): ...


# queue looper


@dataclass(kw_only=True)
class QueueLooper(Looper[int]):
    @override
    async def core(self) -> None:
        await super().core()
        if not self.empty():
            _ = self.get_left_nowait()



from __future__ import annotations

from asyncio import Event, Queue, run, sleep, timeout
from collections import deque
from contextlib import asynccontextmanager
from dataclasses import dataclass, field
from itertools import chain
from re import search
from typing import TYPE_CHECKING, Any, ClassVar, Self, override

from hypothesis import Phase, assume, given, settings
from hypothesis.strategies import (
    DataObject,
    booleans,
    data,
    integers,
    lists,
    none,
    permutations,
    sampled_from,
)
from pytest import LogCaptureFixture, approx, mark, param, raises

from tests.conftest import IS_CI
from tests.test_asyncio_classes.loopers import (
    _BACKOFF,
    _FREQ,
    _REL,
    CountingLooper,
    CountingLooperError,
    LooperWithCounterMixin,
    LooperWithCounterMixins,
    MultipleSubLoopers,
    Outer2CountingLooper,
    OuterCountingLooper,
    QueueLooper,
    assert_looper_full,
    assert_looper_stats,
)
from utilities.asyncio import (
    EnhancedQueue,
    EnhancedTaskGroup,
    Looper,
    UniquePriorityQueue,
    UniqueQueue,
    _LooperNoTaskError,
    get_event,
    get_items,
    get_items_nowait,
    put_items,
    put_items_nowait,
    sleep_dur,
    sleep_max_dur,
    sleep_until,
    sleep_until_rounded,
    stream_command,
    timeout_dur,
)
from utilities.dataclasses import replace_non_sentinel
from utilities.datetime import MILLISECOND, datetime_duration_to_timedelta, get_now
from utilities.functions import get_class_name
from utilities.hypothesis import sentinels, text_ascii
from utilities.iterables import one, unique_everseen
from utilities.pytest import skipif_windows
from utilities.sentinel import Sentinel, sentinel
from utilities.timer import Timer

if TYPE_CHECKING:
    import datetime as dt
    from collections.abc import AsyncIterator

    from utilities.types import Duration, MaybeCallableEvent


class TestEnhancedQueue:
    @given(
        xs=lists(integers()),
        wait=booleans(),
        put_all=booleans(),
        get_reverse=booleans(),
    )
    async def test_left(
        self, *, xs: list[int], wait: int, put_all: bool, get_reverse: bool
    ) -> None:
        _ = assume(not ((len(xs) == 0) and wait))
        deq: deque[int] = deque()
        for x in xs:
            deq.appendleft(x)
        queue: EnhancedQueue[int] = EnhancedQueue()
        if put_all:
            if wait:
                await queue.put_left(*xs)
            else:
                queue.put_left_nowait(*xs)
        else:
            for i, x in enumerate(xs, start=1):
                if wait:
                    await queue.put_left(x)
                else:
                    queue.put_left_nowait(x)
                assert queue.qsize() == i
        assert list(deq) == xs[::-1]
        if wait:
            res = await queue.get_all(reverse=get_reverse)
        else:
            res = queue.get_all_nowait(reverse=get_reverse)
        expected = xs if get_reverse else xs[::-1]
        assert res == expected

    @given(
        xs=lists(integers()),
        wait=booleans(),
        put_all=booleans(),
        get_reverse=booleans(),
    )
    async def test_right(
        self, *, xs: list[int], wait: int, put_all: bool, get_reverse: bool
    ) -> None:
        _ = assume(not ((len(xs) == 0) and wait))
        deq: deque[int] = deque()
        for x in xs:
            deq.append(x)
        queue: EnhancedQueue[int] = EnhancedQueue()
        if put_all:
            if wait:
                await queue.put_right(*xs)
            else:
                queue.put_right_nowait(*xs)
            assert queue.qsize() == len(xs)
        else:
            for i, x in enumerate(xs, start=1):
                if wait:
                    await queue.put_right(x)
                else:
                    queue.put_right_nowait(x)
                assert queue.qsize() == i
        assert list(deq) == xs
        if wait:
            res = await queue.get_all(reverse=get_reverse)
        else:
            res = queue.get_all_nowait(reverse=get_reverse)
        expected = xs[::-1] if get_reverse else xs
        assert res == expected


class TestEnhancedTaskGroup:
    async def test_create_task_context_coroutine(self) -> None:
        flag: bool = False

        @asynccontextmanager
        async def yield_true() -> AsyncIterator[None]:
            nonlocal flag
            try:
                flag = True
                yield
            finally:
                flag = False

        assert not flag
        async with EnhancedTaskGroup(timeout=0.1) as tg:
            _ = tg.create_task_context(yield_true())
            await sleep(0.05)
            assert flag
        assert not flag

    async def test_create_task_context_looper(self) -> None:
        looper = CountingLooper().replace(timeout=1.0)
        assert looper._core_attempts == 0
        async with EnhancedTaskGroup(timeout=0.1) as tg:
            assert looper._core_attempts == 0
            _ = tg.create_task_context(looper)
            await sleep(0.05)
        assert looper._core_attempts >= 1

    async def test_max_tasks_disabled(self) -> None:
        with Timer() as timer:
            async with EnhancedTaskGroup() as tg:
                for _ in range(10):
                    _ = tg.create_task(sleep(0.01))
        assert timer <= 0.05

    async def test_max_tasks_enabled(self) -> None:
        with Timer() as timer:
            async with EnhancedTaskGroup(max_tasks=2) as tg:
                for _ in range(10):
                    _ = tg.create_task(sleep(0.01))
        assert timer >= 0.05

    async def test_timeout_pass(self) -> None:
        async with EnhancedTaskGroup(timeout=0.2) as tg:
            _ = tg.create_task(sleep(0.1))

    async def test_timeout_fail(self) -> None:
        with raises(ExceptionGroup) as exc_info:
            async with EnhancedTaskGroup(timeout=0.05) as tg:
                _ = tg.create_task(sleep(0.1))
        assert len(exc_info.value.exceptions) == 1
        error = one(exc_info.value.exceptions)
        assert isinstance(error, TimeoutError)

    async def test_custom_error(self) -> None:
        class CustomError(Exception): ...

        with raises(ExceptionGroup) as exc_info:
            async with EnhancedTaskGroup(timeout=0.05, error=CustomError) as tg:
                _ = tg.create_task(sleep(0.1))
        assert len(exc_info.value.exceptions) == 1
        error = one(exc_info.value.exceptions)
        assert isinstance(error, CustomError)


class TestGetEvent:
    def test_event(self) -> None:
        event = Event()
        assert get_event(event=event) is event

    @given(event=none() | sentinels())
    def test_none_or_sentinel(self, *, event: None | Sentinel) -> None:
        assert get_event(event=event) is event

    def test_replace_non_sentinel(self) -> None:
        @dataclass(kw_only=True, slots=True)
        class Example:
            event: Event = field(default_factory=Event)

            def replace(
                self, *, event: MaybeCallableEvent | Sentinel = sentinel
            ) -> Self:
                return replace_non_sentinel(self, event=get_event(event=event))

        event1, event2, event3 = Event(), Event(), Event()
        obj = Example(event=event1)
        assert obj.event is event1
        assert obj.replace().event is event1
        assert obj.replace(event=event2).event is event2
        assert obj.replace(event=lambda: event3).event is event3

    def test_callable(self) -> None:
        event = Event()
        assert get_event(event=lambda: event) is event


class TestGetItems:
    @given(
        xs=lists(integers(), min_size=1),
        max_size=integers(1, 10) | none(),
        wait=booleans(),
    )
    async def test_main(
        self, *, xs: list[int], max_size: int | None, wait: bool
    ) -> None:
        queue: Queue[int] = Queue()
        put_items_nowait(xs, queue)
        if wait:
            result = await get_items(queue, max_size=max_size)
        else:
            result = get_items_nowait(queue, max_size=max_size)
        assert result == xs[:max_size]


class TestLooper:
    _restart_min_elapsed: ClassVar[dt.timedelta] = datetime_duration_to_timedelta(
        (0.8 if IS_CI else 1.0) * _BACKOFF
    )
    _restart_max_elapsed: ClassVar[dt.timedelta] = datetime_duration_to_timedelta(
        (1.2 if IS_CI else 1.0) * _FREQ
    )
    skip_sleep_if_failure_cases: ClassVar[list[Any]] = [
        param(True, ""),
        param(False, "; sleeping for .*"),
    ]

    async def test_main_with_nothing(self) -> None:
        looper = CountingLooper()
        async with looper:
            ...
        self._assert_stats_no_runs(looper, entries=1, stops=1)

    async def test_auto_start(self) -> None:
        looper = CountingLooper(auto_start=True)
        with raises(TimeoutError):
            async with timeout(1.0), looper:
                ...
        self._assert_stats_full(looper)

    async def test_auto_start_and_timeout(self) -> None:
        looper = CountingLooper(auto_start=True, timeout=1.0)
        async with looper:
            ...
        self._assert_stats_full(looper, stops=1)

    async def test_await_without_task(self) -> None:
        looper = CountingLooper()
        with raises(_LooperNoTaskError, match=".* has no running task"):
            await looper
        self._assert_stats_no_runs(looper)

    async def test_context_manager_already_entered(
        self, *, caplog: LogCaptureFixture
    ) -> None:
        looper = CountingLooper(timeout=1.0)
        async with looper, looper:
            ...
        _ = one(m for m in caplog.messages if search(": already entered$", m))

    def test_empty(self) -> None:
        looper = CountingLooper()
        assert looper.empty()
        looper.put_left_nowait(None)
        self._assert_stats_no_runs(looper)
        assert not looper.empty()

    async def test_empty_upon_exit(self) -> None:
        looper = QueueLooper(freq=0.05, empty_upon_exit=True)
        looper.put_right_nowait(0)
        assert not looper.empty()
        async with timeout(1.0), looper:
            ...
        self._assert_stats_no_runs(looper, entries=1, stops=1)
        assert looper.empty()

    async def test_explicit_start(self) -> None:
        looper = CountingLooper()
        with raises(TimeoutError):
            async with timeout(1.0), looper:
                await looper
        self._assert_stats_full(looper, stops=1)

    def test_get_all_nowait(self) -> None:
        looper = CountingLooper()
        looper.put_left_nowait(None)
        items = looper.get_all_nowait()
        assert items == [None]
        self._assert_stats_no_runs(looper)

    async def test_initialize_already_initializing(
        self, *, caplog: LogCaptureFixture
    ) -> None:
        class Example(CountingLooper):
            @override
            async def _initialize_core(self) -> None:
                if self._initialization_attempts == 1:
                    _ = await super().initialize()
                await super()._initialize_core()

        looper = Example()
        _ = await looper.initialize()
        _ = one(m for m in caplog.messages if search(": already initializing$", m))

    @mark.parametrize(("skip_sleep_if_failure", "extra"), skip_sleep_if_failure_cases)
    async def test_initialize_failure(
        self, *, skip_sleep_if_failure: bool, extra: str, caplog: LogCaptureFixture
    ) -> None:
        class Example(CountingLooper):
            @override
            async def _initialize_core(self) -> None:
                if self._initialization_attempts == 1:
                    raise CountingLooperError
                await super()._initialize_core()

        looper = Example()
        _ = await looper.initialize(skip_sleep_if_failure=skip_sleep_if_failure)
        pattern = rf": encountered {get_class_name(CountingLooperError)}\(\) whilst initializing{extra}$"
        _ = one(m for m in caplog.messages if search(pattern, m))

    def test_len_and_qsize(self) -> None:
        looper = CountingLooper()
        assert len(looper) == looper.qsize() == 0
        looper.put_left_nowait(None)
        assert len(looper) == looper.qsize() == 1
        self._assert_stats_no_runs(looper)

    @mark.parametrize("counter_auto_start", [param(True), param(False)])
    async def test_mixin(self, *, counter_auto_start: bool) -> None:
        looper = LooperWithCounterMixin(
            auto_start=True, timeout=1.0, counter_auto_start=counter_auto_start
        )
        if counter_auto_start:
            with raises(TimeoutError):
                async with timeout(1.0), looper:
                    ...
            self._assert_stats_half(looper._counter)
        else:
            async with looper:
                ...
            self._assert_stats_half(looper._counter, stops=1)

    @mark.parametrize("counter_auto_start", [param(True), param(False)])
    async def test_mixin_in_task_group(self, *, counter_auto_start: bool) -> None:
        looper = LooperWithCounterMixin(
            auto_start=True, timeout=1.0, counter_auto_start=counter_auto_start
        )
        with raises(ExceptionGroup) as exc_info:
            async with EnhancedTaskGroup(timeout=looper.timeout) as tg:
                _ = tg.create_task_context(looper)
        error = one(exc_info.value.exceptions)
        assert isinstance(error, TimeoutError)
        self._assert_stats_half(looper._counter)

    @mark.parametrize("counter1_auto_start", [param(True), param(False)])
    @mark.parametrize("counter2_auto_start", [param(True), param(False)])
    async def test_mixins(
        self, *, counter1_auto_start: bool, counter2_auto_start: bool
    ) -> None:
        looper = LooperWithCounterMixins(
            auto_start=True,
            timeout=1.0,
            counter1_auto_start=counter1_auto_start,
            counter2_auto_start=counter2_auto_start,
        )
        match counter1_auto_start, counter2_auto_start:
            case _, True:
                with raises(TimeoutError):
                    async with timeout(1.0), looper:
                        ...
                assert_looper_full(looper)
                self._assert_stats_no_runs(looper._counter1)
                self._assert_stats_third(looper._counter2)
            case True, False:
                with raises(TimeoutError):
                    async with timeout(1.0), looper:
                        ...
                assert_looper_full(looper)
                self._assert_stats_half(looper._counter1)
                self._assert_stats_third(looper._counter2)
            case False, False:
                async with looper:
                    ...
                assert_looper_full(looper, stops=1)
                self._assert_stats_half(looper._counter1, stops=1)
                self._assert_stats_third(looper._counter2, stops=1)

    @mark.parametrize("counter_auto_start", [param(True), param(False)])
    async def test_mixins_in_task_group(self, *, counter_auto_start: bool) -> None:
        looper1 = LooperWithCounterMixin(
            auto_start=True, timeout=1.0, counter_auto_start=counter_auto_start
        )
        looper2 = LooperWithCounterMixin(
            auto_start=True, timeout=1.0, counter_auto_start=counter_auto_start
        )
        with raises(ExceptionGroup) as exc_info:  # noqa: PT012
            async with EnhancedTaskGroup(timeout=1.0) as tg:
                _ = tg.create_task_context(looper1)
                _ = tg.create_task_context(looper2)
        errors = exc_info.value.exceptions
        assert 1 <= len(errors) <= 2
        for error in errors:
            assert isinstance(error, TimeoutError)
        self._assert_stats_half(looper1._counter)
        self._assert_stats_half(looper2._counter)

    def test_replace(self) -> None:
        looper = CountingLooper().replace(freq=10.0)
        assert looper.freq == 10.0

    async def test_request_back_off(self) -> None:
        class Example(CountingLooper):
            @override
            async def core(self) -> None:
                await super().core()
                if (self._initialization_attempts >= 2) and (
                    self.count >= (self.max_count / 2)
                ):
                    self.request_back_off()

        looper = Example(auto_start=True, timeout=1.0)
        async with looper:
            ...
        assert_looper_stats(
            looper,
            entries=1,
            core_successes=23,
            core_failures=2,
            initialization_successes=3,
            tear_down_successes=2,
            restart_successes=2,
            stops=1,
        )

    async def test_request_back_off_already_requested(
        self, *, caplog: LogCaptureFixture
    ) -> None:
        looper = CountingLooper()
        for _ in range(2):
            looper.request_back_off()
        _ = one(
            m for m in caplog.messages if search(r": already requested back off$", m)
        )

    async def test_request_restart(self) -> None:
        class Example(CountingLooper):
            @override
            async def core(self) -> None:
                await super().core()
                if (self._initialization_attempts >= 2) and (
                    self.count >= (self.max_count / 2)
                ):
                    self.request_restart()

        looper = Example(auto_start=True, timeout=1.0)
        async with looper:
            ...
        assert_looper_stats(
            looper,
            entries=1,
            core_successes=35,
            core_failures=1,
            initialization_successes=7,
            tear_down_successes=6,
            restart_successes=6,
            stops=1,
        )

    def test_request_restart_already_requested(
        self, *, caplog: LogCaptureFixture
    ) -> None:
        looper = CountingLooper()
        for _ in range(2):
            looper.request_restart()
        _ = one(
            m for m in caplog.messages if search(r": already requested restart$", m)
        )

    async def test_request_stop(self) -> None:
        class Example(CountingLooper):
            @override
            async def core(self) -> None:
                await super().core()
                if (self._initialization_attempts >= 2) and (
                    self.count >= (self.max_count / 2)
                ):
                    self.request_stop()

        looper = Example(auto_start=True, timeout=1.0)
        async with looper:
            ...
        assert_looper_stats(
            looper,
            entries=1,
            core_successes=14,
            core_failures=1,
            initialization_successes=2,
            tear_down_successes=1,
            restart_successes=1,
            stops=1,
        )

    def test_request_stop_already_requested(self, *, caplog: LogCaptureFixture) -> None:
        looper = CountingLooper()
        for _ in range(2):
            looper.request_stop()
        _ = one(m for m in caplog.messages if search(r": already requested stop$", m))

    async def test_request_stop_when_empty(self) -> None:
        class Example(CountingLooper):
            @override
            async def core(self) -> None:
                await super().core()
                if self.empty():
                    await self.stop()
                match self.qsize() % 2 == 0:
                    case True:
                        _ = self.get_left_nowait()
                    case False:
                        _ = self.get_right_nowait()
                self.request_stop_when_empty()

        looper = Example(auto_start=True, timeout=1.0)
        for i in range(25):
            match i % 2 == 0:
                case True:
                    looper.put_left_nowait(i)
                case False:
                    looper.put_right_nowait(i)
        async with looper:
            ...
        assert_looper_stats(
            looper,
            entries=1,
            core_successes=25,
            core_failures=2,
            initialization_successes=3,
            tear_down_successes=2,
            restart_successes=2,
            stops=1,
        )

    def test_request_stop_when_empty_already_requested(
        self, *, caplog: LogCaptureFixture
    ) -> None:
        looper = CountingLooper()
        for _ in range(2):
            looper.request_stop_when_empty()
        _ = one(
            m
            for m in caplog.messages
            if search(r": already requested stop when empty$", m)
        )

    async def test_restart_failure_during_initialization(
        self, *, caplog: LogCaptureFixture
    ) -> None:
        class Example(CountingLooper):
            @override
            async def _initialize_core(self) -> None:
                if self._initialization_attempts == 1:
                    raise CountingLooperError
                await super()._initialize_core()

        looper = Example()
        with Timer() as timer:
            await looper.restart()
        assert timer.timedelta >= self._restart_min_elapsed
        pattern = rf": encountered {get_class_name(CountingLooperError)}\(\) whilst restarting \(initialize\); sleeping for .*\.\.\.$"
        _ = one(m for m in caplog.messages if search(pattern, m))

    async def test_restart_failure_during_tear_down(
        self, *, caplog: LogCaptureFixture
    ) -> None:
        class Example(CountingLooper):
            @override
            async def _tear_down_core(self) -> None:
                if self._tear_down_attempts == 1:
                    raise CountingLooperError
                await super()._tear_down_core()

        looper = Example()
        with Timer() as timer:
            await looper.restart()
        assert timer.timedelta >= self._restart_min_elapsed
        pattern = rf": encountered {get_class_name(CountingLooperError)}\(\) whilst restarting \(tear down\); sleeping for .*\.\.\.$"
        _ = one(m for m in caplog.messages if search(pattern, m))

    async def test_restart_failure_during_tear_down_and_initialization(
        self, *, caplog: LogCaptureFixture
    ) -> None:
        class Example(CountingLooper):
            @override
            async def _initialize_core(self) -> None:
                if self._initialization_attempts == 1:
                    raise CountingLooperError
                await super()._initialize_core()

            @override
            async def _tear_down_core(self) -> None:
                if self._tear_down_attempts == 1:
                    raise CountingLooperError
                await super()._tear_down_core()

        looper = Example()
        with Timer() as timer:
            await looper.restart()
        assert timer.timedelta >= self._restart_min_elapsed
        pattern = rf": encountered {get_class_name(CountingLooperError)}\(\) \(tear down\) and then {get_class_name(CountingLooperError)}\(\) \(initialization\) whilst restarting; sleeping for .*\.\.\.$"
        _ = one(m for m in caplog.messages if search(pattern, m))

    async def test_restart_success(self, *, caplog: LogCaptureFixture) -> None:
        looper = CountingLooper()
        with Timer() as timer:
            await looper.restart()
        assert timer.timedelta <= self._restart_max_elapsed
        pattern = r": finished restarting$"
        _ = one(m for m in caplog.messages if search(pattern, m))

    @mark.parametrize("n", [param(0), param(1), param(2)])
    async def test_run_until_empty(self, *, n: int) -> None:
        looper = QueueLooper(freq=0.05)
        looper.put_right_nowait(*range(n))
        async with timeout(1.0), looper:
            await looper.run_until_empty()
        assert looper.empty()

    @mark.parametrize("inner_auto_start", [param(True), param(False)])
    async def test_sub_looper_one(self, *, inner_auto_start: bool) -> None:
        looper = OuterCountingLooper(
            auto_start=True, timeout=1.0, inner_auto_start=inner_auto_start
        )
        if inner_auto_start:
            with raises(TimeoutError):
                async with timeout(1.0), looper:
                    ...
            self._assert_stats_full(looper)
            self._assert_stats_half(looper.inner)
        else:
            async with looper:
                ...
            self._assert_stats_full(looper, stops=1)
            self._assert_stats_half(looper.inner, stops=1)

    @mark.parametrize("inner1_auto_start", [param(True), param(False)])
    @mark.parametrize("inner2_auto_start", [param(True), param(False)])
    async def test_sub_loopers_multiple(
        self, *, inner1_auto_start: bool, inner2_auto_start: bool
    ) -> None:
        looper = MultipleSubLoopers(
            auto_start=True,
            timeout=1.0,
            inner1_auto_start=inner1_auto_start,
            inner2_auto_start=inner2_auto_start,
        )
        match inner1_auto_start, inner2_auto_start:
            case True, _:
                with raises(TimeoutError):
                    async with timeout(1.0), looper:
                        ...
                self._assert_stats_full(looper)
                self._assert_stats_half(looper.inner1)
                self._assert_stats_no_runs(looper.inner2)
            case False, True:
                with raises(TimeoutError):
                    async with timeout(1.0), looper:
                        ...
                self._assert_stats_full(looper)
                self._assert_stats_half(looper.inner1)
                self._assert_stats_third(looper.inner2)
            case False, False:
                async with looper:
                    ...
                self._assert_stats_full(looper, stops=1)
                self._assert_stats_half(looper.inner1, stops=1)
                self._assert_stats_third(looper.inner2, stops=1)

    @mark.parametrize("middle_auto_start", [param(True), param(False)])
    @mark.parametrize("inner_auto_start", [param(True), param(False)])
    async def test_sub_loopers_nested(
        self, *, middle_auto_start: bool, inner_auto_start: bool
    ) -> None:
        looper = Outer2CountingLooper(
            auto_start=True,
            timeout=1.0,
            middle_auto_start=middle_auto_start,
            inner_auto_start=inner_auto_start,
        )
        match middle_auto_start, inner_auto_start:
            case False, False:
                async with looper:
                    ...
                self._assert_stats_full(looper, stops=1)
                self._assert_stats_half(looper.middle, stops=1)
                self._assert_stats_quarter(looper.middle.inner, stops=1)
            case _, _:
                with raises(TimeoutError):
                    async with timeout(1.0), looper:
                        ...
                self._assert_stats_full(looper)
                self._assert_stats_half(looper.middle)
                self._assert_stats_quarter(looper.middle.inner)

    async def test_tear_down_already_tearing_down(
        self, *, caplog: LogCaptureFixture
    ) -> None:
        class Example(CountingLooper):
            @override
            async def _tear_down_core(self) -> None:
                if self._tear_down_attempts == 1:
                    _ = await super().tear_down()
                await super()._tear_down_core()

        looper = Example()
        _ = await looper.tear_down()
        _ = one(m for m in caplog.messages if search(": already tearing down$", m))

    @mark.parametrize(("skip_sleep_if_failure", "extra"), skip_sleep_if_failure_cases)
    async def test_tear_down_failure(
        self, *, skip_sleep_if_failure: bool, extra: str, caplog: LogCaptureFixture
    ) -> None:
        class Example(CountingLooper):
            @override
            async def _tear_down_core(self) -> None:
                if self._tear_down_attempts == 1:
                    raise CountingLooperError
                await super()._tear_down_core()

        looper = Example()
        _ = await looper.tear_down(skip_sleep_if_failure=skip_sleep_if_failure)
        pattern = rf": encountered {get_class_name(CountingLooperError)}\(\) whilst tearing down{extra}$"
        _ = one(m for m in caplog.messages if search(pattern, m))

    async def test_timeout(self) -> None:
        looper = CountingLooper(timeout=1.0)
        with Timer() as timer:
            async with looper:
                await looper
        assert float(timer) == approx(1.0, rel=_REL)
        self._assert_stats_full(looper, stops=1)

    def test_with_auto_start(self) -> None:
        looper = CountingLooper()
        assert not looper.auto_start
        assert looper.with_auto_start.auto_start

    def _assert_stats_no_runs(
        self,
        looper: Looper[Any],
        /,
        *,
        entries: int = 0,
        stops: int = 0,
        rel: float = _REL,
    ) -> None:
        assert_looper_stats(looper, entries=entries, stops=stops, rel=rel)

    def _assert_stats_full(
        self, looper: Looper[Any], /, *, stops: int = 0, rel: float = _REL
    ) -> None:
        assert_looper_stats(
            looper,
            entries=1,
            core_successes=47,
            core_failures=5,
            initialization_successes=6,
            tear_down_successes=5,
            restart_successes=5,
            stops=stops,
            rel=rel,
        )

    def _assert_stats_half(
        self, looper: Looper[Any], /, *, stops: int = 0, rel: float = _REL
    ) -> None:
        assert_looper_stats(
            looper,
            entries=1,
            core_successes=56,
            core_failures=14,
            initialization_successes=15,
            tear_down_successes=14,
            restart_successes=14,
            stops=stops,
            rel=rel,
        )

    def _assert_stats_third(
        self, looper: Looper[Any], /, *, stops: int = 0, rel: float = _REL
    ) -> None:
        assert_looper_stats(
            looper,
            entries=1,
            core_successes=48,
            core_failures=24,
            initialization_successes=25,
            tear_down_successes=24,
            restart_successes=24,
            stops=stops,
            rel=rel,
        )

    def _assert_stats_quarter(
        self, looper: Looper[Any], /, *, stops: int = 0, rel: float = _REL
    ) -> None:
        assert_looper_stats(
            looper,
            entries=1,
            core_successes=35,
            core_failures=35,
            initialization_successes=36,
            tear_down_successes=35,
            restart_successes=35,
            stops=stops,
            rel=rel,
        )


class TestPutItems:
    @given(xs=lists(integers(), min_size=1), wait=booleans())
    async def test_main(self, *, xs: list[int], wait: bool) -> None:
        queue: Queue[int] = Queue()
        if wait:
            put_items_nowait(xs, queue)
        else:
            await put_items(xs, queue)
        result: list[int] = []
        while not queue.empty():
            result.append(await queue.get())
        assert result == xs


class TestUniquePriorityQueue:
    @given(data=data(), texts=lists(text_ascii(min_size=1), min_size=1, unique=True))
    async def test_main(self, *, data: DataObject, texts: list[str]) -> None:
        items = list(enumerate(texts))
        extra = data.draw(lists(sampled_from(items)))
        items_use = data.draw(permutations(list(chain(items, extra))))
        queue: UniquePriorityQueue[int, str] = UniquePriorityQueue()
        assert queue._set == set()
        for item in items_use:
            await queue.put(item)
        assert queue._set == set(texts)
        result = await get_items(queue)
        assert result == items
        assert queue._set == set()


class TestUniqueQueue:
    @given(x=lists(integers(), min_size=1))
    async def test_main(self, *, x: list[int]) -> None:
        queue: UniqueQueue[int] = UniqueQueue()
        assert queue._set == set()
        for x_i in x:
            await queue.put(x_i)
        assert queue._set == set(x)
        result = await get_items(queue)
        expected = list(unique_everseen(x))
        assert result == expected
        assert queue._set == set()


class TestSleepDur:
    @given(duration=sampled_from([0.1, 10 * MILLISECOND]))
    @settings(phases={Phase.generate})
    async def test_main(self, *, duration: Duration) -> None:
        with Timer() as timer:
            await sleep_dur(duration=duration)
        assert timer <= datetime_duration_to_timedelta(2 * duration)

    async def test_none(self) -> None:
        with Timer() as timer:
            await sleep_dur()
        assert timer <= 0.01


class TestSleepMaxDur:
    @given(duration=sampled_from([0.1, 10 * MILLISECOND]))
    @settings(phases={Phase.generate})
    async def test_main(self, *, duration: Duration) -> None:
        with Timer() as timer:
            await sleep_max_dur(duration=duration)
        assert timer <= datetime_duration_to_timedelta(2 * duration)

    async def test_none(self) -> None:
        with Timer() as timer:
            await sleep_max_dur()
        assert timer <= 0.01


class TestSleepUntil:
    async def test_main(self) -> None:
        await sleep_until(get_now() + 10 * MILLISECOND)


class TestSleepUntilRounded:
    async def test_main(self) -> None:
        await sleep_until_rounded(10 * MILLISECOND)


class TestStreamCommand:
    @skipif_windows
    async def test_main(self) -> None:
        output = await stream_command(
            'echo "stdout message" && sleep 0.1 && echo "stderr message" >&2'
        )
        await sleep(0.01)
        assert output.return_code == 0
        assert output.stdout == "stdout message\n"
        assert output.stderr == "stderr message\n"

    @skipif_windows
    async def test_error(self) -> None:
        output = await stream_command("this-is-an-error")
        await sleep(0.01)
        assert output.return_code == 127
        assert output.stdout == ""
        assert search(
            r"^/bin/sh: (1: )?this-is-an-error: (command )?not found$", output.stderr
        )


class TestTimeoutDur:
    async def test_pass(self) -> None:
        async with timeout_dur(duration=0.2):
            await sleep(0.1)

    async def test_fail(self) -> None:
        with raises(TimeoutError):
            async with timeout_dur(duration=0.05):
                await sleep(0.1)

    async def test_custom_error(self) -> None:
        class CustomError(Exception): ...

        with raises(CustomError):
            async with timeout_dur(duration=0.05, error=CustomError):
                await sleep(0.1)


if __name__ == "__main__":
    _ = run(
        stream_command('echo "stdout message" && sleep 2 && echo "stderr message" >&2')
    )



from __future__ import annotations

from typing import Any

from hypothesis import given
from hypothesis.strategies import sampled_from

from utilities.reprlib import (
    get_call_args_mapping,
    get_repr,
    get_repr_and_class,
    yield_call_args_repr,
    yield_mapping_repr,
)


class TestGetCallArgsMapping:
    def test_main(self) -> None:
        mapping = get_call_args_mapping(1, 2, 3, x=4, y=5, z=6)
        expected = {
            "args[0]": 1,
            "args[1]": 2,
            "args[2]": 3,
            "kwargs[x]": 4,
            "kwargs[y]": 5,
            "kwargs[z]": 6,
        }
        assert mapping == expected


class TestGetRepr:
    @given(
        case=sampled_from([
            (None, "None"),
            (0, "0"),
            (
                list(range(21)),
                "[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, ... +1]",
            ),
        ])
    )
    def test_main(self, *, case: tuple[Any, str]) -> None:
        obj, expected = case
        result = get_repr(obj)
        assert result == expected


class TestGetReprAndClass:
    @given(
        case=sampled_from([
            (None, "Object 'None' of type 'NoneType'"),
            (0, "Object '0' of type 'int'"),
        ])
    )
    def test_main(self, *, case: tuple[Any, str]) -> None:
        obj, expected = case
        result = get_repr_and_class(obj)
        assert result == expected


class TestYieldCallArgsRepr:
    def test_main(self) -> None:
        lines = list(
            yield_call_args_repr(1, 2, list(range(100)), d=4, e=5, f=list(range(100)))
        )
        expected = [
            "args[0] = 1",
            "args[1] = 2",
            "args[2] = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, ... +80]",
            "kwargs[d] = 4",
            "kwargs[e] = 5",
            "kwargs[f] = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, ... +80]",
        ]
        assert lines == expected


class TestYieldMappingRepr:
    def test_main(self) -> None:
        mapping = {"a": 1, "b": 2, "c": 3, "d": list(range(100))}
        lines = list(yield_mapping_repr(mapping))
        expected = [
            "a = 1",
            "b = 2",
            "c = 3",
            "d = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, ... +80]",
        ]
        assert lines == expected



from __future__ import annotations

from typing import cast

from inflect import Word, engine

_ENGINE = engine()


def counted_noun(num: int, noun: str, /) -> str:
    """Construct a counted noun."""
    word = cast("Word", noun)
    sin_or_plu = _ENGINE.plural_noun(word, count=num)
    return f"{num} {sin_or_plu}"


__all__ = ["counted_noun"]



from __future__ import annotations

from collections.abc import Callable
from contextlib import contextmanager, suppress
from dataclasses import dataclass
from itertools import chain
from os import chdir
from os.path import expandvars
from pathlib import Path
from re import IGNORECASE, search
from subprocess import PIPE, CalledProcessError, check_output
from typing import TYPE_CHECKING, assert_never, overload, override

from utilities.sentinel import Sentinel, sentinel

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence

    from utilities.types import MaybeCallablePathLike, PathLike

PWD = Path.cwd()


def ensure_suffix(path: PathLike, suffix: str, /) -> Path:
    """Ensure a path has a given suffix."""
    path = Path(path)
    parts = path.name.split(".")
    parts = list(chain([parts[0]], (f".{p}" for p in parts[1:])))
    if (len(parts) == 0) or (parts[-1] != suffix):
        parts.append(suffix)
    name = "".join(parts)
    return path.with_name(name)


##


def expand_path(path: PathLike, /) -> Path:
    """Expand a path."""
    path = str(path)
    path = expandvars(path)
    return Path(path).expanduser()


##


@overload
def get_path(*, path: MaybeCallablePathLike | None) -> Path: ...
@overload
def get_path(*, path: Sentinel) -> Sentinel: ...
def get_path(
    *, path: MaybeCallablePathLike | None | Sentinel = sentinel
) -> Path | None | Sentinel:
    """Get the path."""
    match path:
        case Path() | Sentinel():
            return path
        case str():
            return Path(path)
        case None:
            return Path.cwd()
        case Callable() as func:
            return get_path(path=func())
        case _ as never:
            assert_never(never)


##


def get_root(*, path: MaybeCallablePathLike | None = None) -> Path:
    """Get the root of a path."""
    path = get_path(path=path)
    try:
        output = check_output(
            ["git", "rev-parse", "--show-toplevel"], stderr=PIPE, cwd=path, text=True
        )
    except CalledProcessError as error:
        # newer versions of git report "Not a git repository", whilst older
        # versions report "not a git repository"
        if not search("fatal: not a git repository", error.stderr, flags=IGNORECASE):
            raise  # pragma: no cover
    else:
        return Path(output.strip("\n"))
    all_paths = list(chain([path], path.parents))
    with suppress(StopIteration):
        return next(
            p for p in all_paths if any(p_i.name == ".envrc" for p_i in p.iterdir())
        )
    raise GetRootError(path=path)


@dataclass(kw_only=True, slots=True)
class GetRootError(Exception):
    path: PathLike

    @override
    def __str__(self) -> str:
        return f"Unable to determine root from {str(self.path)!r}"


##


def list_dir(path: PathLike, /) -> Sequence[Path]:
    """List the contents of a directory."""
    return sorted(Path(path).iterdir())


##


@contextmanager
def temp_cwd(path: PathLike, /) -> Iterator[None]:
    """Context manager with temporary current working directory set."""
    prev = Path.cwd()
    chdir(path)
    try:
        yield
    finally:
        chdir(prev)


__all__ = ["PWD", "ensure_suffix", "expand_path", "get_path", "list_dir", "temp_cwd"]



from __future__ import annotations

import datetime as dt
from datetime import timezone
from logging import DEBUG
from re import escape
from typing import TYPE_CHECKING
from zoneinfo import ZoneInfo

from hypothesis import assume, example, given
from hypothesis.strategies import (
    DataObject,
    data,
    dates,
    datetimes,
    integers,
    just,
    sampled_from,
    timedeltas,
    times,
    timezones,
)
from pytest import raises
from whenever import DateTimeDelta, ZonedDateTime

from tests.conftest import SKIPIF_CI_AND_WINDOWS
from utilities.datetime import (
    _MICROSECONDS_PER_DAY,
    _MICROSECONDS_PER_SECOND,
    DAY,
    MICROSECOND,
    SECOND,
    maybe_sub_pct_y,
    parse_two_digit_year,
    serialize_compact,
)
from utilities.hypothesis import (
    assume_does_not_raise,
    datetime_durations,
    plain_datetimes,
    timedeltas_2w,
    zoned_datetimes,
)
from utilities.tzdata import HongKong
from utilities.whenever import (
    MAX_SERIALIZABLE_TIMEDELTA,
    MIN_SERIALIZABLE_TIMEDELTA,
    EnsureDateError,
    EnsureDateTimeError,
    EnsureDurationError,
    EnsurePlainDateTimeError,
    EnsureTimeError,
    EnsureZonedDateTimeError,
    ParseDateError,
    ParseDateTimeError,
    ParseDurationError,
    ParsePlainDateTimeError,
    ParseTimeError,
    ParseZonedDateTimeError,
    SerializeDurationError,
    SerializePlainDateTimeError,
    SerializeTimeDeltaError,
    SerializeZonedDateTimeError,
    WheneverLogRecord,
    _CheckValidZonedDateTimeUnequalError,
    _EnsureTimedeltaNanosecondError,
    _EnsureTimedeltaParseError,
    _ParseTimedeltaNanosecondError,
    _ParseTimedeltaParseError,
    _to_datetime_delta,
    _ToDateTimeDeltaError,
    check_valid_zoned_datetime,
    ensure_date,
    ensure_datetime,
    ensure_duration,
    ensure_plain_datetime,
    ensure_time,
    ensure_timedelta,
    ensure_zoned_datetime,
    parse_date,
    parse_datetime,
    parse_duration,
    parse_plain_datetime,
    parse_time,
    parse_timedelta,
    parse_zoned_datetime,
    serialize_date,
    serialize_datetime,
    serialize_duration,
    serialize_plain_datetime,
    serialize_time,
    serialize_timedelta,
    serialize_zoned_datetime,
)
from utilities.zoneinfo import UTC, get_time_zone_name

if TYPE_CHECKING:
    from utilities.types import Duration


_TIMEDELTA_MICROSECONDS = int(1e18) * MICROSECOND
_TIMEDELTA_OVERFLOW = dt.timedelta(days=106751991, seconds=14454, microseconds=775808)


@SKIPIF_CI_AND_WINDOWS
class TestCheckValidZonedDateTime:
    @given(
        datetime=sampled_from([
            dt.datetime(1951, 4, 1, 3, tzinfo=HongKong),
            dt.datetime(1951, 4, 1, 5, tzinfo=HongKong),
        ])
    )
    def test_main(self, *, datetime: dt.datetime) -> None:
        check_valid_zoned_datetime(datetime)

    def test_error(self) -> None:
        datetime = dt.datetime(1951, 4, 1, 4, tzinfo=HongKong)
        with raises(
            _CheckValidZonedDateTimeUnequalError,
            match=escape(
                "Zoned datetime must be valid; got 1951-04-01 04:00:00+08:00 != 1951-04-01 05:00:00+09:00"
            ),
        ):
            check_valid_zoned_datetime(datetime)


class TestSerializeAndParseDate:
    @given(date=dates())
    def test_main(self, *, date: dt.date) -> None:
        serialized = serialize_date(date)
        result = parse_date(serialized)
        assert result == date

    @given(year=integers(0, 99), date=dates())
    def test_yymmdd(self, *, year: int, date: dt.date) -> None:
        with assume_does_not_raise(ValueError):
            date_use = date.replace(parse_two_digit_year(year))
        serialized = date_use.strftime("%y%m%d")
        result = parse_date(serialized)
        assert result == date_use

    @given(date=dates())
    def test_yyyymmdd(self, *, date: dt.date) -> None:
        serialized = serialize_compact(date)
        result = parse_date(serialized)
        assert result == date

    def test_error_parse(self) -> None:
        with raises(ParseDateError, match="Unable to parse date; got 'invalid'"):
            _ = parse_date("invalid")

    @given(data=data(), date=dates())
    def test_ensure(self, *, data: DataObject, date: dt.date) -> None:
        str_or_value = data.draw(sampled_from([date, serialize_date(date)]))
        result = ensure_date(str_or_value)
        assert result == date

    def test_error_ensure(self) -> None:
        with raises(EnsureDateError, match="Unable to ensure date; got 'invalid'"):
            _ = ensure_date("invalid")


class TestSerializeAndParseDateTime:
    @given(
        datetime=plain_datetimes()
        | zoned_datetimes(time_zone=timezones() | just(dt.UTC), valid=True)
    )
    @SKIPIF_CI_AND_WINDOWS
    def test_main(self, *, datetime: dt.datetime) -> None:
        serialized = serialize_datetime(datetime)
        result = parse_datetime(serialized)
        assert result == datetime

    def test_error_parse(self) -> None:
        with raises(
            ParseDateTimeError, match="Unable to parse datetime; got 'invalid'"
        ):
            _ = parse_datetime("invalid")

    @given(
        data=data(),
        datetime=plain_datetimes()
        | zoned_datetimes(time_zone=timezones() | just(dt.UTC), valid=True),
    )
    @SKIPIF_CI_AND_WINDOWS
    def test_ensure(self, *, data: DataObject, datetime: dt.datetime) -> None:
        str_or_value = data.draw(sampled_from([datetime, serialize_datetime(datetime)]))
        result = ensure_datetime(str_or_value)
        assert result == datetime

    def test_error_ensure(self) -> None:
        with raises(
            EnsureDateTimeError, match="Unable to ensure datetime; got 'invalid'"
        ):
            _ = ensure_datetime("invalid")


class TestSerializeAndParseDuration:
    @given(duration=datetime_durations())
    def test_main(self, *, duration: Duration) -> None:
        with assume_does_not_raise(SerializeDurationError):
            serialized = serialize_duration(duration)
        with assume_does_not_raise(ParseDurationError):
            result = parse_duration(serialized)
        assert result == duration

    def test_error_parse(self) -> None:
        with raises(
            ParseDurationError, match="Unable to parse duration; got 'invalid'"
        ):
            _ = parse_duration("invalid")

    @given(duration=sampled_from([_TIMEDELTA_MICROSECONDS, _TIMEDELTA_OVERFLOW]))
    def test_error_serialize(self, *, duration: Duration) -> None:
        with raises(
            SerializeDurationError, match="Unable to serialize duration; got .*"
        ):
            _ = serialize_duration(duration)

    @given(data=data(), duration=datetime_durations(two_way=True))
    def test_ensure(self, *, data: DataObject, duration: Duration) -> None:
        str_or_value = data.draw(sampled_from([duration, serialize_duration(duration)]))
        result = ensure_duration(str_or_value)
        assert result == duration

    def test_error_ensure(self) -> None:
        with raises(
            EnsureDurationError, match="Unable to ensure duration; got 'invalid'"
        ):
            _ = ensure_duration("invalid")


class TestSerializeAndParsePlainDateTime:
    @given(datetime=datetimes())
    def test_main(self, *, datetime: dt.datetime) -> None:
        serialized = serialize_plain_datetime(datetime)
        result = parse_plain_datetime(serialized)
        assert result == datetime

    @given(datetime=plain_datetimes(round_="standard", timedelta=SECOND))
    def test_compact_no_microseconds(self, *, datetime: dt.datetime) -> None:
        assert datetime.microsecond == 0
        serialized = datetime.strftime(maybe_sub_pct_y("%Y%m%dT%H%M%S"))
        result = parse_plain_datetime(serialized)
        assert result == datetime

    @given(datetime=datetimes())
    def test_compact_with_microseconds(self, *, datetime: dt.datetime) -> None:
        _ = assume(datetime.microsecond != 0)
        serialized = datetime.strftime(maybe_sub_pct_y("%Y%m%dT%H%M%S.%f"))
        result = parse_plain_datetime(serialized)
        assert result == datetime

    def test_error_parse(self) -> None:
        with raises(
            ParsePlainDateTimeError,
            match="Unable to parse plain datetime; got 'invalid'",
        ):
            _ = parse_plain_datetime("invalid")

    def test_error_serialize(self) -> None:
        datetime = dt.datetime(2000, 1, 1, tzinfo=UTC)
        with raises(
            SerializePlainDateTimeError,
            match="Unable to serialize plain datetime; got .*",
        ):
            _ = serialize_plain_datetime(datetime)

    @given(data=data(), datetime=datetimes())
    def test_ensure(self, *, data: DataObject, datetime: dt.datetime) -> None:
        str_or_value = data.draw(
            sampled_from([datetime, serialize_plain_datetime(datetime)])
        )
        result = ensure_plain_datetime(str_or_value)
        assert result == datetime

    def test_error_ensure(self) -> None:
        with raises(
            EnsurePlainDateTimeError,
            match="Unable to ensure plain datetime; got 'invalid'",
        ):
            _ = ensure_plain_datetime("invalid")


class TestSerializeAndParseTime:
    @given(time=times())
    def test_main(self, *, time: dt.time) -> None:
        serialized = serialize_time(time)
        result = parse_time(serialized)
        assert result == time

    def test_error_parse(self) -> None:
        with raises(ParseTimeError, match="Unable to parse time; got 'invalid'"):
            _ = parse_time("invalid")

    @given(data=data(), time=times())
    def test_ensure(self, *, data: DataObject, time: dt.time) -> None:
        str_or_value = data.draw(sampled_from([time, serialize_time(time)]))
        result = ensure_time(str_or_value)
        assert result == time

    def test_error_ensure(self) -> None:
        with raises(EnsureTimeError, match="Unable to ensure time; got 'invalid'"):
            _ = ensure_time("invalid")


class TestSerializeAndParseTimedelta:
    @given(
        timedelta=timedeltas(
            min_value=MIN_SERIALIZABLE_TIMEDELTA, max_value=MAX_SERIALIZABLE_TIMEDELTA
        )
    )
    @example(timedelta=int(1e6) * DAY)
    @example(timedelta=MICROSECOND)
    @example(timedelta=-DAY)
    @example(timedelta=-DAY + MICROSECOND)
    def test_main(self, *, timedelta: dt.timedelta) -> None:
        serialized = serialize_timedelta(timedelta)
        result = parse_timedelta(serialized)
        assert result == timedelta

    @given(timedelta=timedeltas(min_value=MICROSECOND))
    def test_min_serializable(self, *, timedelta: dt.timedelta) -> None:
        _ = serialize_timedelta(MIN_SERIALIZABLE_TIMEDELTA)
        with assume_does_not_raise(OverflowError):
            offset = MIN_SERIALIZABLE_TIMEDELTA - timedelta
        with raises(SerializeTimeDeltaError):
            _ = serialize_timedelta(offset)

    @given(timedelta=timedeltas(min_value=MICROSECOND))
    def test_max_serializable(self, *, timedelta: dt.timedelta) -> None:
        _ = serialize_timedelta(MAX_SERIALIZABLE_TIMEDELTA)
        with assume_does_not_raise(OverflowError):
            offset = MAX_SERIALIZABLE_TIMEDELTA + timedelta
        with raises(SerializeTimeDeltaError):
            _ = serialize_timedelta(offset)

    def test_error_parse(self) -> None:
        with raises(
            _ParseTimedeltaParseError, match="Unable to parse timedelta; got 'invalid'"
        ):
            _ = parse_timedelta("invalid")

    def test_error_parse_nano_seconds(self) -> None:
        with raises(
            _ParseTimedeltaNanosecondError,
            match="Unable to parse timedelta; got 333 nanoseconds",
        ):
            _ = parse_timedelta("PT0.111222333S")

    @given(timedelta=sampled_from([_TIMEDELTA_MICROSECONDS, _TIMEDELTA_OVERFLOW]))
    def test_error_serialize(self, *, timedelta: dt.timedelta) -> None:
        with raises(
            SerializeTimeDeltaError, match="Unable to serialize timedelta; got .*"
        ):
            _ = serialize_timedelta(timedelta)

    @given(data=data(), timedelta=timedeltas_2w())
    def test_ensure(self, *, data: DataObject, timedelta: dt.timedelta) -> None:
        str_or_value = data.draw(
            sampled_from([timedelta, serialize_timedelta(timedelta)])
        )
        result = ensure_timedelta(str_or_value)
        assert result == timedelta

    def test_error_ensure(self) -> None:
        with raises(
            _EnsureTimedeltaParseError,
            match="Unable to ensure timedelta; got 'invalid'",
        ):
            _ = ensure_timedelta("invalid")

    def test_error_ensure_nano_seconds(self) -> None:
        with raises(
            _EnsureTimedeltaNanosecondError,
            match="Unable to ensure timedelta; got 333 nanoseconds",
        ):
            _ = ensure_timedelta("PT0.111222333S")


class TestSerializeAndParseZonedDateTime:
    @given(datetime=zoned_datetimes(valid=True))
    @SKIPIF_CI_AND_WINDOWS
    def test_main(self, *, datetime: dt.datetime) -> None:
        serialized = serialize_zoned_datetime(datetime)
        result = parse_zoned_datetime(serialized)
        assert result == datetime

    @given(
        datetime=zoned_datetimes(
            time_zone=just(dt.UTC), round_="standard", timedelta=SECOND, valid=True
        )
    )
    @SKIPIF_CI_AND_WINDOWS
    def test_compact_no_microseconds(self, *, datetime: dt.datetime) -> None:
        assert datetime.microsecond == 0
        part1 = datetime.strftime(maybe_sub_pct_y("%Y%m%dT%H%M%S"))
        assert isinstance(datetime.tzinfo, ZoneInfo | timezone)
        part2 = get_time_zone_name(datetime.tzinfo)
        serialized = f"{part1}[{part2}]"
        result = parse_zoned_datetime(serialized)
        assert result == datetime

    @given(datetime=zoned_datetimes(time_zone=just(dt.UTC), valid=True))
    @SKIPIF_CI_AND_WINDOWS
    def test_compact_with_microseconds(self, *, datetime: dt.datetime) -> None:
        _ = assume(datetime.microsecond != 0)
        part1 = datetime.strftime(maybe_sub_pct_y("%Y%m%dT%H%M%S.%f"))
        assert isinstance(datetime.tzinfo, ZoneInfo | timezone)
        part2 = get_time_zone_name(datetime.tzinfo)
        serialized = f"{part1}[{part2}]"
        result = parse_zoned_datetime(serialized)
        assert result == datetime

    @given(datetime=plain_datetimes())
    @SKIPIF_CI_AND_WINDOWS
    def test_utc(self, *, datetime: dt.datetime) -> None:
        datetime1, datetime2 = [datetime.astimezone(tz) for tz in [UTC, dt.UTC]]
        serialized1, serialized2 = map(serialize_zoned_datetime, [datetime1, datetime2])
        assert serialized1 == serialized2
        parsed = parse_zoned_datetime(serialized1)
        assert abs(parsed - datetime1) <= SECOND
        assert abs(parsed - datetime2) <= SECOND
        assert parsed.tzinfo is UTC

    def test_error_parse(self) -> None:
        with raises(
            ParseZonedDateTimeError,
            match="Unable to parse zoned datetime; got 'invalid'",
        ):
            _ = parse_zoned_datetime("invalid")

    def test_error_serialize(self) -> None:
        datetime = dt.datetime(2000, 1, 1).astimezone(None)
        with raises(
            SerializeZonedDateTimeError,
            match="Unable to serialize zoned datetime; got 2000-01-01 00:00:00",
        ):
            _ = serialize_zoned_datetime(datetime)

    @given(
        data=data(),
        datetime=zoned_datetimes(time_zone=timezones() | just(dt.UTC), valid=True),
    )
    @SKIPIF_CI_AND_WINDOWS
    def test_ensure(self, *, data: DataObject, datetime: dt.datetime) -> None:
        str_or_value = data.draw(
            sampled_from([datetime, serialize_zoned_datetime(datetime)])
        )
        result = ensure_zoned_datetime(str_or_value)
        assert result == datetime

    def test_error_ensure(self) -> None:
        with raises(
            EnsureZonedDateTimeError,
            match="Unable to ensure zoned datetime; got 'invalid'",
        ):
            _ = ensure_zoned_datetime("invalid")


class TestToDateTimeDelta:
    @given(days=integers(), microseconds=integers())
    def test_main(self, *, days: int, microseconds: int) -> None:
        with assume_does_not_raise(OverflowError):
            timedelta = dt.timedelta(days=days, microseconds=microseconds)
        init_total_micro = _MICROSECONDS_PER_DAY * days + microseconds
        with assume_does_not_raise(_ToDateTimeDeltaError):
            result = _to_datetime_delta(timedelta)
        comp_month, comp_day, comp_sec, comp_nano = result.in_months_days_secs_nanos()
        assert comp_month == 0
        comp_micro, remainder = divmod(comp_nano, 1000)
        assert remainder == 0
        result_total_micro = (
            _MICROSECONDS_PER_DAY * comp_day
            + _MICROSECONDS_PER_SECOND * comp_sec
            + comp_micro
        )
        assert init_total_micro == result_total_micro

    def test_mixed_sign(self) -> None:
        timedelta = dt.timedelta(days=-1, seconds=1)
        result = _to_datetime_delta(timedelta)
        expected = DateTimeDelta(seconds=timedelta.total_seconds())
        assert result == expected

    def test_close_to_overflow(self) -> None:
        timedelta = dt.timedelta(days=104250, microseconds=1)
        result = _to_datetime_delta(timedelta)
        expected = DateTimeDelta(days=104250, microseconds=1)
        assert result == expected

    @given(timedelta=sampled_from([_TIMEDELTA_MICROSECONDS, _TIMEDELTA_OVERFLOW]))
    def test_error(self, *, timedelta: dt.timedelta) -> None:
        with raises(
            _ToDateTimeDeltaError, match="Unable to create DateTimeDelta; got .*"
        ):
            _ = _to_datetime_delta(timedelta)


class TestWheneverLogRecord:
    def test_init(self) -> None:
        _ = WheneverLogRecord("name", DEBUG, "pathname", 0, None, None, None)

    def test_get_length(self) -> None:
        assert isinstance(WheneverLogRecord._get_length(), int)

    def test_get_now(self) -> None:
        assert isinstance(WheneverLogRecord._get_now(), ZonedDateTime)

    def test_get_time_zone(self) -> None:
        assert isinstance(WheneverLogRecord._get_time_zone(), ZoneInfo)

    def test_get_time_zone_key(self) -> None:
        assert isinstance(WheneverLogRecord._get_time_zone_key(), str)
